# The NONMEM Model class
from __future__ import annotations

import re
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Tuple

from pharmpy.deps import pandas as pd
from pharmpy.deps import sympy
from pharmpy.internals.fs.path import path_absolute, path_relative_to
from pharmpy.internals.immutable import frozenmapping
from pharmpy.model import Assignment, DataInfo, EstimationSteps
from pharmpy.model import Model as BaseModel
from pharmpy.model import NormalDistribution, Parameter, Parameters, RandomVariables, Statements
from pharmpy.model.model import ModelInternals, compare_before_after_params, update_datainfo
from pharmpy.modeling.write_csv import write_csv

from .nmtran_parser import NMTranControlStream, NMTranParser
from .parsing import (
    convert_dvs,
    parse_datainfo,
    parse_dataset,
    parse_description,
    parse_estimation_steps,
    parse_initial_individual_estimates,
    parse_parameters,
    parse_statements,
    parse_value_type,
)
from .update import (
    abbr_translation,
    create_name_map,
    update_ccontra,
    update_description,
    update_estimation,
    update_initial_individual_estimates,
    update_input,
    update_name_of_tables,
    update_random_variables,
    update_sizes,
    update_statements,
    update_thetas,
)


@dataclass(frozen=True)
class NONMEMModelInternals(ModelInternals):
    control_stream: NMTranControlStream
    old_name: str
    old_description: str
    old_estimation_steps: EstimationSteps
    old_observation_transformation: sympy.Symbol
    old_parameters: Parameters
    old_random_variables: RandomVariables
    old_statements: Statements
    old_initial_individual_estimates: Optional[pd.DataFrame]
    old_datainfo: DataInfo
    old_dependent_variables: dict
    compartment_map: Optional[Dict[str, int]]
    name_map: Dict[str, str]


def convert_model(model):
    """Convert any model into a NONMEM model"""
    if isinstance(model, Model):
        return model

    code = '$PROBLEM\n'
    code += (
        '$INPUT '
        + ' '.join(
            f'{column.name}=DROP' if column.drop else column.name for column in model.datainfo
        )
        + '\n'
    )
    code += '$DATA file.csv IGNORE=@\n'
    if model.statements.ode_system is None:
        code += '$PRED\nY=X\n'
    else:
        code += '$SUBROUTINES ADVAN1 TRANS2\n'
        code += '$PK\nY=X\n'
        code += '$ERROR\nA=B\n'
    nm_model = parse_model(code, dataset=model.dataset)
    assert isinstance(nm_model, Model)
    nm_model._datainfo = model.datainfo
    nm_model._parameters = model.parameters
    nm_model._dataset = model.dataset
    nm_model._estimation_steps = model.estimation_steps
    nm_model._initial_individual_estimates = model.initial_individual_estimates
    new_obs_trans = frozenmapping({dv: dv for dv in model.dependent_variables.keys()})
    internals = nm_model.internals.replace(old_parameters=Parameters())
    nm_model = nm_model.replace(
        name=model.name,
        value_type=model.value_type,
        observation_transformation=new_obs_trans,
        dependent_variables=model.dependent_variables,
        random_variables=model.random_variables,
        statements=model.statements,
        description=model.description,
        internals=internals,
    )
    nm_model = nm_model.update_source()
    if model.statements.ode_system:
        nm_model = nm_model.replace(
            internals=nm_model.internals.replace(
                compartment_map={
                    name: i
                    for i, name in enumerate(model.statements.ode_system.compartment_names, start=1)
                }
            )
        )
    return nm_model


class Model(BaseModel):
    def __init__(
        self,
        internals: NONMEMModelInternals,
        **kwargs,
    ):
        super().__init__(
            **kwargs,
        )
        assert isinstance(internals, NONMEMModelInternals)
        self._internals = internals

    def update_source(self):
        """Update the source

        path - path to modelfile
        nofiles - Set to not write any files (i.e. dataset, phi input etc)
        """
        model = self
        if (
            model.initial_individual_estimates
            is not model.internals.old_initial_individual_estimates
        ):
            cs = update_initial_individual_estimates(model, path='DUMMYPATH', nofiles=True)
            model = model.replace(
                internals=model.internals.replace(
                    control_stream=cs,
                    old_initial_individual_estimates=model.initial_individual_estimates,
                )
            )

        if not model.random_variables.etas:
            omega = Parameter('DUMMYOMEGA', init=0, fix=True)
            eta = NormalDistribution.create('eta_dummy', 'iiv', 0, omega.symbol)
            statement = Assignment(sympy.Symbol('DUMMYETA'), sympy.Symbol(eta.names[0]))
            model = model.replace(
                statements=statement + model.statements,
                random_variables=model.random_variables + eta,
                parameters=Parameters.create(list(model.parameters) + [omega]),
            )

        control_stream = update_random_variables(
            model, model.internals.old_random_variables, model._random_variables
        )

        control_stream = update_thetas(
            model, control_stream, model.internals.old_parameters, model._parameters
        )

        model = model.replace(
            internals=model.internals.replace(
                old_parameters=model._parameters,
                old_random_variables=model._random_variables,
                control_stream=control_stream,
            )
        )

        # Parameters that needs to be renamed in statements
        trans = create_name_map(model)

        # RVs that need $ABBR (either has proper names or have been renumbered
        rv_trans = {}
        i = 1
        for dist in model._random_variables.etas:
            for name in dist.names:
                nonmem_pattern = re.match(r'ETA[_(]([0-9]+)\)*', name)
                if not nonmem_pattern:
                    rv_trans[name] = f'ETA({i})'
                elif nonmem_pattern.group(1) != str(i):
                    rv_trans[name] = f'ETA({i})'
                i += 1

        if model._random_variables.etas.names != ['eta_dummy']:
            model, abbr_map = abbr_translation(model, rv_trans)
            trans = {key: value for key, value in trans.items() if key not in abbr_map.values()}

        trans = {sympy.Symbol(key): sympy.Symbol(value) for key, value in trans.items()}
        model, updated_dataset = update_statements(
            model, model.internals.old_statements, model._statements, trans
        )
        # model = update_dependent_variables(model, trans)

        cs = model.internals.control_stream

        if model.dataset is None:
            pass
        elif (
            updated_dataset
            or (model.datainfo.path is None and model.dataset is not None)
            or model.datainfo != model.internals.old_datainfo
            or model.datainfo.path != model.internals.old_datainfo.path
        ):
            data_record = control_stream.get_records('DATA')[0]
            label = model.datainfo.names[0]
            newdata = data_record.set_ignore_character_from_header(label)
            cs = update_input(cs, model)

            # Remove IGNORE/ACCEPT. Could do diff between old dataset and find simple
            # IGNOREs to add i.e. for filter out certain ID.
            newdata = newdata.remove_ignore().remove_accept()
            if (
                model.datainfo.path is None
                or (model.datainfo.path is None and model.dataset is not None)
                or updated_dataset
            ):
                newdata = newdata.set_filename('DUMMYPATH')

            cs = cs.replace_records([data_record], [newdata])

        cs = update_sizes(cs, model)
        cs = update_estimation(cs, model)
        cs = update_description(cs, model.internals.old_description, model.description)

        if model._name != model.internals.old_name:
            cs = update_name_of_tables(cs, model._name)

        new_internals = model.internals.replace(
            control_stream=cs,
            old_description=model._description,
            old_estimation_steps=model._estimation_steps,
            old_datainfo=model._datainfo,
            old_statements=model._statements,
        )
        model = model.replace(internals=new_internals)

        return model

    def write_files(self, path=None, force=False):
        model = self.update_source()

        etas_record = model.internals.control_stream.get_records('ETAS')
        if etas_record and str(etas_record[0].path) == 'DUMMYPATH':
            cs = update_initial_individual_estimates(model, path)
            model = model.replace(
                internals=model.internals.replace(
                    control_stream=cs,
                )
            )

        replace_dict = {}

        data_record = model.internals.control_stream.get_records('DATA')[0]
        if data_record.filename == 'DUMMYPATH' or force:
            datapath = model.datainfo.path
            if datapath is None:
                dir_path = Path(model.name + ".csv") if path is None else path.parent
                model = write_csv(model, path=dir_path, force=force)
                datapath = model.datainfo.path
                replace_dict['datainfo'] = model.datainfo.replace(path=datapath)
            assert (
                not datapath.exists() or datapath.is_file()
            ), f'input path change, but no file exists at target {str(datapath)}'
            parent_path = Path.cwd() if path is None else path.parent
            try:
                filename = str(path_relative_to(parent_path, datapath))
            except ValueError:
                # NOTE: If parent path and datapath are in different drives absolute path
                # needs to be used
                filename = str(path_absolute(datapath))
                warnings.warn('Cannot resolve relative path, falling back to absolute path')
            newdata = data_record.set_filename(filename)
            newcs = model.internals.control_stream.replace_records([data_record], [newdata])
            replace_dict['internals'] = model.internals.replace(control_stream=newcs)

        if model.observation_transformation != model.internals.old_observation_transformation:
            update_ccontra(model, path, force)

        if replace_dict:
            return model.replace(**replace_dict)
        else:
            return model

    @property
    def model_code(self):
        return str(self.internals.control_stream)

    def read_raw_dataset(self, parse_columns: Tuple[str, ...] = ()):
        return parse_dataset(
            self.datainfo, self.internals.control_stream, raw=True, parse_columns=parse_columns
        )


def parse_model(
    code: str, path: Optional[Path] = None, dataset: Optional[pd.DataFrame] = None, **_
):
    parser = NMTranParser()
    if path is None:
        name = 'run1'
        filename_extension = '.ctl'
    else:
        name = path.stem
        filename_extension = path.suffix

    control_stream = parser.parse(code)
    di = parse_datainfo(control_stream, path)
    old_datainfo = di
    # FIXME: Temporary workaround remove when changing constructor
    # NOTE: Inputting the dataset is needed for IV models when using convert_model, since it needs to read the
    # RATE column to decide dosing, meaning it needs the dataset before parsing statements
    if dataset is not None:
        di = update_datainfo(di.replace(path=None), dataset)

    try:
        dataset = parse_dataset(di, control_stream, raw=False)
    except FileNotFoundError:
        pass

    statements, comp_map = parse_statements(di, dataset, control_stream)
    statements, dependent_variables, obs_trans = convert_dvs(statements, control_stream)

    parameters, rvs, name_map = parse_parameters(control_stream, statements)

    subs_map = {sympy.Symbol(key): sympy.Symbol(val) for key, val in name_map.items()}
    statements = statements.subs(subs_map)

    # FIXME: Handle by creation of new model object
    if not rvs.validate_parameters(parameters.inits):
        nearest = rvs.nearest_valid_parameters(parameters.inits)
        before, after = compare_before_after_params(parameters.inits, nearest)
        warnings.warn(
            f"Adjusting initial estimates to create positive semidefinite "
            f"omega/sigma matrices.\nBefore adjusting:  {before}.\n"
            f"After adjusting: {after}"
        )
        parameters = parameters.set_initial_estimates(nearest)

    estimation_steps = parse_estimation_steps(control_stream, rvs)

    description = parse_description(control_stream)

    init_etas = parse_initial_individual_estimates(
        control_stream, name_map, None if path is None else path.parent
    )

    value_type = parse_value_type(control_stream, statements)

    internals = NONMEMModelInternals(
        control_stream=control_stream,
        old_name=name,
        old_description=description,
        old_estimation_steps=estimation_steps,
        old_observation_transformation=obs_trans,
        old_parameters=parameters,
        old_random_variables=rvs,
        old_statements=statements,
        old_initial_individual_estimates=init_etas,
        old_datainfo=old_datainfo,
        old_dependent_variables=dependent_variables,
        compartment_map=comp_map,
        name_map=name_map,
    )

    return Model(
        name=name,
        parameters=parameters,
        random_variables=rvs,
        statements=statements,
        dataset=dataset,
        datainfo=di,
        dependent_variables=dependent_variables,
        observation_transformation=obs_trans,
        estimation_steps=estimation_steps,
        parent_model=None,
        initial_individual_estimates=init_etas,
        filename_extension=filename_extension,
        value_type=value_type,
        description=description,
        internals=internals,
    )



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pharmpy.model
from pharmpy.internals.code_generator import CodeGenerator
from pharmpy.model.model import ModelInternals
from pharmpy.modeling import drop_columns, get_evid, translate_nmtran_time, update_inits

from .error_model import res_error_term
from .ini import add_eta, add_sigma, add_theta
from .model_block import add_bio_lag, add_ode, add_statements
from .sanity_checks import check_model


def convert_model(
    model: pharmpy.model.Model,
    skip_check: bool = False,
    updated_estimates: bool = False,
) -> pharmpy.model.Model:
    """
    Convert a NONMEM model into an nlmixr model

    Parameters
    ----------
    model : pharmpy.model.Model
        A NONMEM pharmpy model object
    skip_check : bool, optional
        Skip determination of error model type. Could speed up conversion. The default is False.

    Returns
    -------
    pharmpy.model.Model
        A model converted to nlmixr format.

    """

    if isinstance(model, Model):
        return model

    if updated_estimates:
        model = update_inits(model, model.modelfit_results.parameter_estimates)

    nlmixr_model = Model(
        internals=NLMIXRModelInternals(),
        parameters=model.parameters,
        random_variables=model.random_variables,
        statements=model.statements,
        dependent_variables=model.dependent_variables,
        estimation_steps=model.estimation_steps,
        filename_extension='.R',
        datainfo=model.datainfo,
        dataset=model.dataset,
        name=model.name,
        description=model.description,
    )

    # Update dataset
    if model.dataset is not None:
        nlmixr_model = translate_nmtran_time(nlmixr_model)
        # FIXME: Dropping columns runs update source which becomes redundant.
        # drop_dropped_columns(nlmixr_model)
        if all(x in nlmixr_model.dataset.columns for x in ["RATE", "DUR"]):
            nlmixr_model = drop_columns(nlmixr_model, ["DUR"])
        nlmixr_model = nlmixr_model.replace(
            datainfo=nlmixr_model.datainfo.replace(path=None),
            dataset=nlmixr_model.dataset.reset_index(drop=True),
        )

        # Add evid
        nlmixr_model = add_evid(nlmixr_model)

    # Check model for warnings regarding data structure or model contents
    nlmixr_model = check_model(nlmixr_model, skip_error_model_check=skip_check)

    nlmixr_model.update_source()

    return nlmixr_model


def create_dataset(cg: CodeGenerator, model: pharmpy.model.Model, path=None) -> None:
    """
    Create dataset for nlmixr

    Parameters
    ----------
    cg : CodeGenerator
        A code object associated with the model.
    model : pharmpy.model.Model
        A pharmpy.model object.
    path : TYPE, optional
        Path to add file to. The default is None.

    """
    dataname = f'{model.name}.csv'
    if path is None:
        path = ""
    path = Path(path) / dataname
    cg.add(f'dataset <- read.csv("{path}")')


def create_ini(cg: CodeGenerator, model: pharmpy.model.Model) -> None:
    """
    Create the nlmixr ini block code

    Parameters
    ----------
    cg : CodeGenerator
        A code object associated with the model.
    model : pharmpy.model.Model
        A pharmpy.model object.

    """
    cg.add('ini({')
    cg.indent()

    add_theta(model, cg)

    add_eta(model, cg)

    add_sigma(model, cg)

    cg.dedent()
    cg.add('})')


def create_model(cg: CodeGenerator, model: pharmpy.model.Model) -> None:
    """
    Create the nlmixr model block code

    Parameters
    ----------
    cg : CodeGenerator
        A code object associated with the model.
    model : pharmpy.model.Model
        A pharmpy.model object.

    """

    cg.add('model({')

    # Add statements before ODEs
    cg.indent()
    if len(model.statements.after_odes) != 0:
        add_statements(model, cg, model.statements.before_odes)

    # Add the ODEs
    cg.add("")
    cg.add("# --- DIFF EQUATIONS ---")
    if model.statements.ode_system:
        add_ode(model, cg)
    cg.add("")

    # Find what kind of error model we are looking at
    dv = list(model.dependent_variables.keys())[0]
    dv_statement = model.statements.find_assignment(dv)

    only_piecewise = False
    if dv_statement.expression.is_Piecewise:
        only_piecewise = True
        dependencies = set()
        res_alias = set()
        for s in model.statements.after_odes:
            if s.symbol == dv:
                if s.expression.is_Piecewise:
                    for value, cond in s.expression.args:
                        if value != dv:
                            dv_term = res_error_term(model, value)
                            dependencies.update(dv_term.dependencies())

                            dv_term.create_res_alias()
                            res_alias.update(dv_term.res_alias)
                else:
                    dv_term = res_error_term(model, s.expression)
                    dependencies.update(dv_term.dependencies())

                    dv_term.create_res_alias()
                    res_alias.update(dv_term.res_alias)
    else:
        dv_term = res_error_term(model, dv_statement.expression)
        dependencies = dv_term.dependencies()
        dv_term.create_res_alias()
        res_alias = dv_term.res_alias

    # Add bioavailability statements
    if model.statements.ode_system is not None:
        add_bio_lag(model, cg, bio=True)
        add_bio_lag(model, cg, lag=True)

    # Add statements after ODEs
    if len(model.statements.after_odes) == 0:
        statements = model.statements
    else:
        statements = model.statements.after_odes
        comp_name_dict = {}
        for name in model.statements.ode_system.compartment_names:
            comp = model.statements.ode_system.find_compartment(name)
            comp_name_dict[comp.amount] = comp.amount.name
        statements = statements.subs(comp_name_dict)
    add_statements(
        model, cg, statements, only_piecewise, dependencies=dependencies, res_alias=res_alias
    )

    cg.dedent()
    cg.add('})')


def create_fit(cg: CodeGenerator, model: pharmpy.model.Model) -> None:
    """
    Create the call to fit for the nlmixr model with appropriate methods and datasets

    Parameters
    ----------
    cg : CodeGenerator
        A code object associated with the model.
    model : pharmpy.model
        A pharmpy.model.Model object.

    """
    # FIXME: Raise error if the method does not match when evaluating
    estimation_steps = model.estimation_steps[0]
    if "fix_eta" in estimation_steps.tool_options:
        fix_eta = True
    else:
        fix_eta = False

    if [s.evaluation for s in model.estimation_steps._steps][0] is True:
        max_eval = 0
    else:
        max_eval = estimation_steps.maximum_evaluations

    method = estimation_steps.method
    interaction = estimation_steps.interaction

    nonmem_method_to_nlmixr = {"FOCE": "foce", "FO": "fo", "SAEM": "saem"}

    if method not in nonmem_method_to_nlmixr.keys():
        nlmixr_method = "focei"
    else:
        nlmixr_method = nonmem_method_to_nlmixr[method]

    if interaction and nlmixr_method != "saem":
        nlmixr_method += "i"

    if max_eval is not None:
        if max_eval == 0 and nlmixr_method not in ["fo", "foi", "foce", "focei"]:
            nlmixr_method = "posthoc"
            cg.add(f'fit <- nlmixr2({model.name}, dataset, est = "{nlmixr_method}"')
        else:
            f = f'fit <- nlmixr2({model.name}, dataset, est = "{nlmixr_method}", '
            if fix_eta:
                f += f'control=foceiControl(maxOuterIterations={max_eval}, maxInnerIterations=0, etaMat = etas))'
            else:
                f += f'control=foceiControl(maxOuterIterations={max_eval}))'
            cg.add(f)
    else:
        cg.add(f'fit <- nlmixr2({model.name}, dataset, est = "{nlmixr_method}")')


def add_evid(model: pharmpy.model.Model) -> pharmpy.model.Model:
    temp_model = model
    if "EVID" not in temp_model.dataset.columns:
        temp_model.dataset["EVID"] = get_evid(temp_model)
    return temp_model


@dataclass(frozen=True)
class NLMIXRModelInternals(ModelInternals):
    src: Optional[str] = None
    path: Optional[Path] = None


class Model(pharmpy.model.Model):
    def __init__(self, **kwargs):
        super().__init__(
            **kwargs,
        )

    def update_source(self):
        cg = CodeGenerator()
        cg.add(f'{self.name} <- function() {{')
        cg.indent()
        create_ini(cg, self)
        create_model(cg, self)
        cg.dedent()
        cg.add('}')
        cg.empty_line()
        create_fit(cg, self)
        # Create lowercase id, time and amount symbols for nlmixr to be able
        # to run
        code = str(cg).replace("AMT", "amt").replace("TIME", "time")
        path = None
        internals = self.internals.replace(src=code, path=path)
        model = self.replace(internals=internals)
        return model

    @property
    def model_code(self):
        model = self.update_source()
        code = model.internals.src
        assert code is not None
        return code



"""
.. highlight:: console

============================
The CLI interface of Pharmpy
============================

Examples
--------

Example argument lines (only preceded by ``pharmpy`` or ``python3 -m pharmpy``)::

    # filter data (as model prescribes) and output to file 'output.csv'
    pharmpy data write run1.mod -o output.csv

Entrypoints and Subcommands
---------------------------

Standard practice of packaging is to define "entrypoints" in the setuptools ``setup.py``. Currently:

.. code-block:: python

  entry_points={
      'console_scripts': [
          'pharmpy           = pharmpy.cli:main',
      ]
  },

Main entrypoint is ``pharmpy`` (which maps :func:`main`), but each subcommand could be given a
shortcutting entrypoint (e.g. :func:`sumo`). This would be accomplished via invoking:

Installing Pharmpy Wheel-build should install all of these to PATH. Then, these are clean targets
for post-install hooks to symlink. Either to pollute the PATH namespace with traditional PsN-styled
"binscripts", as::

    /usr/bin/execute               ->  /usr/bin/pharmpy execute

Or, to keep multiple versions when upgrading Pharmpy.

On Logging
----------

As a CLI we are at the top, interacting with the user. All of Pharmpy will get messages
filtered through here (loggers organize themselves in a hierarchy based on their period-split names).

.. warning::
    Loggers **mustn't** be configured if imported as library (e.g. set handlers and such),
    since that should bubble up into the lap of whomever is at the top (of the Python
    period-separated namespace).

Definitions
===========
"""

import argparse
import pydoc
import sys
import warnings
from collections import namedtuple
from pathlib import Path
from textwrap import dedent

import pharmpy
from pharmpy.internals.fs.path import path_absolute

from .deps import pandas as pd

formatter = argparse.ArgumentDefaultsHelpFormatter


def warnings_formatter(message, *args, **kwargs):
    return f'Warning: {message}\n'


warnings.formatwarning = warnings_formatter


def error(exception):
    """Raise the exception with no traceback printed

    Used for non-recoverables in CLI. Exceptions giving traceback are bugs!
    """

    def exc_only_hook(exception_type, exception, traceback):
        print(f'{exception_type.__name__}: {exception}')

    sys.excepthook = exc_only_hook
    raise exception


def format_keyval_pairs(data_dict, sort=True, right_just=False):
    """Formats lines from *data_dict*."""

    if not data_dict:
        return []

    key_width = max(len(field) for field in data_dict.keys())
    if sort:
        data_dict = dict(sorted(data_dict.items()))
    if right_just:
        line_format = '  %%%ds\t%%s' % key_width
    else:
        line_format = '%%-%ds\t%%s' % key_width

    lines = []
    for key, values in data_dict.items():
        if isinstance(values, str):
            values = values.splitlines()
        keys = [key] + [''] * (len(values) - 1)
        for k, v in zip(keys, values):
            lines += [line_format % (k, v)]

    return lines


def run_bootstrap(args):
    from pharmpy.tools import run_bootstrap

    run_bootstrap(args.model, args.model.modelfit_results, resamples=args.samples)


def run_execute(args):
    from pharmpy.tools import fit

    fit(args.models)


def run_modelsearch(args):
    from pharmpy.tools import run_tool

    run_tool(
        'modelsearch',
        args.mfl,
        args.algorithm,
        rank_type=args.rank_type,
        cutoff=args.cutoff,
        iiv_strategy=args.iiv_strategy,
        results=args.model.modelfit_results,
        model=args.model,
        path=args.path,
    )


def run_iivsearch(args):
    from pharmpy.tools import run_tool

    run_tool(
        'iivsearch',
        args.algorithm,
        iiv_strategy=args.iiv_strategy,
        rank_type=args.rank_type,
        cutoff=args.cutoff,
        model=args.model,
        results=args.model.modelfit_results,
        path=args.path,
    )


def run_iovsearch(args):
    from pharmpy.tools import run_tool

    run_tool(
        'iovsearch',
        column=args.column,
        list_of_parameters=args.list_of_parameters,
        rank_type=args.rank_type,
        cutoff=args.cutoff,
        distribution=args.distribution,
        model=args.model,
        results=args.model.modelfit_results,
        path=args.path,
    )


def run_covsearch(args):
    from pharmpy.tools import run_tool

    run_tool(
        'covsearch',
        effects=args.effects,
        p_forward=args.p_forward,
        max_steps=args.max_steps,
        algorithm=args.algorithm,
        results=args.model.modelfit_results,
        model=args.model,
        path=args.path,
    )


def run_ruvsearch(args):
    from pharmpy.tools import run_tool

    run_tool(
        'ruvsearch',
        model=args.model,
        groups=args.groups,
        p_value=args.p_value,
        skip=args.skip,
        path=args.path,
    )


def run_allometry(args):
    from pharmpy.tools import run_tool

    run_tool(
        'allometry',
        model=args.model,
        results=args.model.modelfit_results,
        allometric_variable=args.allometric_variable,
        reference_value=args.reference_value,
        parameters=args.parameters,
        initials=args.initials,
        lower_bounds=args.lower_bounds,
        upper_bounds=args.upper_bounds,
        fixed=not args.non_fixed,
        path=args.path,
    )


def run_estmethod(args):
    from pharmpy.tools import run_tool

    try:
        methods = args.methods.split(" ")
    except AttributeError:
        methods = args.methods
    try:
        solvers = args.solvers.split(" ")
    except AttributeError:
        solvers = args.solvers

    run_tool(
        'estmethod',
        args.algorithm,
        methods=methods,
        solvers=solvers,
        model=args.model,
        path=args.path,
    )


def run_amd(args):
    from pharmpy.tools import run_amd

    run_amd(
        args.input_path,
        modeltype=args.modeltype,
        administration=args.administration,
        cl_init=args.cl_init,
        vc_init=args.vc_init,
        mat_init=args.mat_init,
        search_space=args.search_space,
        lloq_method=args.lloq_method,
        lloq_limit=args.lloq_limit,
        order=args.order,
        allometric_variable=args.allometric_variable,
        occasion=args.occasion,
        path=args.path,
        resume=args.resume,
    )


def run_retries(args):
    from pharmpy.tools import run_tool

    run_tool(
        'retries',
        model=args.model,
        results=args.model.modelfit_results,  # This feels wierd to do
        number_of_candidates=args.number_of_candidates,
    )


def data_write(args):
    """Subcommand to write a dataset."""
    try:
        from pharmpy.modeling import write_csv

        # If no output_file supplied will use name of df
        path = write_csv(args.model, path=args.output_file, force=args.force)
        print(f'Dataset written to {path}')
    except OSError as e:
        error(e)


def data_print(args):
    """Subcommand to print a dataset to stdout via pager"""
    try:
        df = args.model_or_dataset.dataset
    except AttributeError:
        df = args.model_or_dataset
    if args.columns:
        try:
            df = df[args.columns]
        except KeyError as e:
            error(e)
    pydoc.pager(df.to_string())


def data_filter(args):
    """Subcommand to filter a dataset"""
    try:
        df = args.model_or_dataset.dataset
    except TypeError:
        df = args.model_or_dataset
    expression = ' '.join(args.expressions)
    try:
        df = df.query(expression)
    except SyntaxError:
        error(SyntaxError(f'Invalid syntax of query: "{expression}"'))
    except BaseException as e:
        error(e)
    write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)


def data_append(args):
    """Subcommand to append a column to a dataset"""
    try:
        df = args.model_or_dataset.dataset
    except TypeError:
        df = args.model_or_dataset
    try:
        df = df.eval(args.expression)
    except SyntaxError:
        error(SyntaxError(f'Invalid syntax of expression: "{args.expression}"'))
    write_model_or_dataset(args.model_or_dataset, df, args.output_file, args.force)


def write_model_or_dataset(model_or_dataset, new_df, path, force):
    """Write model or dataset to output_path or using default names"""
    if isinstance(model_or_dataset, pd.DataFrame):
        # Is a dataset
        try:
            # If no output_file supplied will use name of df
            if force or not path.is_file():
                path = new_df.to_csv(path, index=False)
                print(f'Dataset written to {path}')
            else:
                error(exception=FileExistsError("Use -f or --force to force an overwrite"))
        except FileExistsError as e:
            error(
                exception=FileExistsError(
                    f'{e.args[0]} Use -f  or --force to ' 'force an overwrite'
                )
            )
    else:
        # Is a model
        model = model_or_dataset
        if new_df is not None and new_df is not model.dataset:
            model = model.replace(dataset=new_df)
        try:
            from pharmpy.modeling import bump_model_number, write_model

            if path:
                if not path.is_dir():
                    model = bump_model_number(model, path)
                write_model(model, path=path, force=force)
            else:
                model = bump_model_number(model, path='.')
                write_model(model, force=force)
        except FileExistsError as e:
            error(FileExistsError(f'{e.args[0]} Use -f or --force to ' 'force an overwrite'))


def data_resample(args):
    """Subcommand to resample a dataset."""
    from pharmpy.modeling import resample_data

    resampler = resample_data(
        args.model_or_dataset,
        args.group,
        resamples=args.resamples,
        stratify=args.stratify,
        replace=args.replace,
        sample_size=args.sample_size,
        name_pattern=f'{args.model_or_dataset.name}_resample_{{}}',
    )
    for resampled_obj, _ in resampler:
        try:
            try:
                from pharmpy.modeling import write_model

                write_model(resampled_obj)
            except AttributeError:
                # FIXME!
                resampled_obj.to_csv()
        except OSError as e:
            error(e)


def data_deidentify(args):
    """Subcommand to deidentify a dataset
    Will be default try to overwrite original file (but won't unless force)
    """
    path = check_input_path(args.dataset)
    dataset = pd.read_csv(path, dtype=str)

    from pharmpy.modeling import deidentify_data

    if args.datecols:
        datecols = args.datecols.split(',')
    else:
        datecols = None
    df = deidentify_data(dataset, id_column=args.idcol, date_columns=datecols)

    if args.output_file:
        output_file = args.output_file
    else:
        output_file = path
    write_model_or_dataset(df, df, output_file, args.force)


def data_reference(args):
    """Subcommand to replace columns with reference values"""
    from pharmpy.modeling import set_reference_values

    d = key_vals(args.colrefs)
    model = args.model
    model = set_reference_values(model, d)
    write_model_or_dataset(model, model.dataset, args.output_file, args.force)


def info(args):
    """Subcommand to print Pharmpy info (and brag a little bit)."""

    import pharmpy

    Install = namedtuple('VersionInfo', ['version', 'authors', 'directory'])
    inst = Install(
        pharmpy.__version__,
        'A list of authors can be found in the AUTHORS.rst',
        str(Path(pharmpy.__file__).resolve().parent),
    )

    lines = format_keyval_pairs(inst._asdict(), right_just=True)
    print('\n'.join(lines))


def model_print(args):
    """Subcommand for formatting/printing model components."""
    lines = []
    for i, model in enumerate(args.models):
        lines += ['[%d/%d] %r' % (i + 1, len(args.models), model.name)]
        dict_ = {}
        try:
            dict_['dataset'] = repr(model.dataset)
        except FileNotFoundError as e:
            dict_['dataset'] = str(e)
        s = repr(model.parameters) + '\n\n'
        dict_['parameters'] = s
        s = repr(model.random_variables) + '\n\n'
        dict_['random_variables'] = s
        s = str(model.statements) + '\n\n'
        dict_['model_statements'] = s
        dict_lines = format_keyval_pairs(dict_, sort=False)
        lines += ['\t%s' % line for line in dict_lines]
    if len(lines) > 24:
        pydoc.pager('\n'.join(lines))
    else:
        print('\n'.join(lines))


def model_sample(args):
    model = args.model
    from pharmpy.modeling import sample_parameters_from_covariance_matrix

    samples = sample_parameters_from_covariance_matrix(
        model,
        model.modelfit_results.parameter_estimates,
        model.modelfit_results.covariance_matrix,
        n=args.samples,
    )
    for row, params in samples.iterrows():
        from pharmpy.modeling import set_initial_estimates, write_model

        set_initial_estimates(model, params)
        model = model.replace(name=f'sample_{row + 1}')
        write_model(model)


def add_covariate_effect(args):
    """Subcommand to add covariate effect to model."""
    from pharmpy.modeling import add_covariate_effect

    model = args.model
    model = add_covariate_effect(model, args.param, args.covariate, args.effect, args.operation)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def model_absorption_rate(args):
    import pharmpy.modeling as modeling

    model = args.model
    if args.order == 'ZO':
        modeling.set_zero_order_absorption(model)
    elif args.order == 'FO':
        modeling.set_first_order_absorption(model)
    elif args.order == 'bolus':
        modeling.set_bolus_absorption(model)
    elif args.order == 'seq-ZO-FO':
        modeling.set_seq_zo_fo_absorption(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_elimination_rate(args):
    model = args.model
    if args.order == 'ZO':
        from pharmpy.modeling import set_zero_order_elimination

        set_zero_order_elimination(model)
    elif args.order == 'FO':
        from pharmpy.modeling import set_first_order_elimination

        set_first_order_elimination(model)
    elif args.order == 'MM':
        from pharmpy.modeling import set_michaelis_menten_elimination

        set_michaelis_menten_elimination(model)
    elif args.order == 'comb-FO-MM':
        from pharmpy.modeling import set_mixed_mm_fo_elimination

        set_mixed_mm_fo_elimination(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_transit_compartments(args):
    from pharmpy.modeling import set_transit_compartments

    model = args.model
    set_transit_compartments(model, args.n)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_peripheral_compartments(args):
    from pharmpy.modeling import add_peripheral_compartment, remove_peripheral_compartment

    model = args.model
    p = model.statements.ode_system.peripheral_compartments
    if len(p) < args.n:
        func = add_peripheral_compartment
    elif len(p) > args.n:
        func = remove_peripheral_compartment
    else:
        return
    iters = abs(len(p) - args.n)
    for _ in range(iters):
        func(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def model_error(args):
    import pharmpy.modeling as modeling

    model = args.model
    if args.error_type == 'none':
        model = modeling.remove_error_model(model)
    elif args.error_model == 'additive':
        model = modeling.set_additive_error_model(model)
    elif args.error_model == 'proportional':
        model = modeling.set_proportional_error_model(model)
    else:  # args.error_model == 'combined':
        model = modeling.set_combined_error_model(model)
    write_model_or_dataset(model, None, path=args.output_file, force=args.force)


def boxcox(args):
    """Subcommand to apply boxcox transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_boxcox

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    model = transform_etas_boxcox(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def tdist(args):
    """Subcommand to apply t-distribution transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_tdist

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    model = transform_etas_tdist(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def john_draper(args):
    """Subcommand to apply John Draper transformation to specified etas of model."""
    from pharmpy.modeling import transform_etas_john_draper

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    model = transform_etas_john_draper(model, etas)
    write_model_or_dataset(model, None, path=args.output_file, force=False)


def add_iiv(args):
    """Subcommand to add new IIVs to model."""
    from pharmpy.modeling import add_iiv

    model = args.model
    model = add_iiv(
        model,
        list_of_parameters=args.param,
        expression=args.expression,
        operation=args.operation,
        eta_names=args.eta_name,
    )

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def add_iov(args):
    """Subcommand to add new IOVs to model."""
    from pharmpy.modeling import add_iov

    model = args.model
    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    try:
        eta_names = args.eta_names.split(" ")
    except AttributeError:
        eta_names = args.eta_names

    model = add_iov(model, args.occ, etas, eta_names)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def create_joint_distribution(args):
    """Subcommand to create full or partial block structures."""
    from pharmpy.modeling import create_joint_distribution

    model = args.model

    try:
        etas = args.etas.split(" ")
    except AttributeError:
        etas = args.etas

    model = create_joint_distribution(
        model,
        etas,
        individual_estimates=model.modelfit_results.individual_estimates
        if model.modelfit_results is not None
        else None,
    )

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def iiv_on_ruv(args):
    """Subcommand to multiply epsilons with exponential etas."""
    from pharmpy.modeling import set_iiv_on_ruv

    model = args.model

    try:
        eps = args.eps.split(" ")
    except AttributeError:
        eps = args.eps

    try:
        eta_names = args.eta_names.split(" ")
    except AttributeError:
        eta_names = args.eta_names

    model = set_iiv_on_ruv(model, list_of_eps=eps, same_eta=args.same_eta, eta_names=eta_names)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def remove_iiv(args):
    """Subcommand to remove IIVs."""
    from pharmpy.modeling import remove_iiv

    model = args.model

    try:
        to_remove = args.to_remove.split(" ")
    except AttributeError:
        to_remove = args.to_remove

    model = remove_iiv(model, to_remove)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def remove_iov(args):
    """Subcommand to remove IOVs."""
    from pharmpy.modeling import remove_iov

    model = args.model
    model = remove_iov(model)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def power_on_ruv(args):
    """Subcommand to apply power effect to RUVs."""
    from pharmpy.modeling import set_power_on_ruv

    model = args.model

    try:
        eps = args.eps.split(" ")
    except AttributeError:
        eps = args.eps

    model = set_power_on_ruv(model, eps)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def update_inits(args):
    """Subcommand to update initial estimates from previous output."""
    from pharmpy.modeling import update_inits

    model = args.model
    update_inits(model, model.modelfit_results.parameter_estimates)

    write_model_or_dataset(model, model.dataset, path=args.output_file, force=False)


def results_bootstrap(args):
    """Subcommand to generate bootstrap results"""
    from pharmpy.tools.bootstrap.results import psn_bootstrap_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    try:
        res = psn_bootstrap_results(args.psn_dir)
    except FileNotFoundError as e:
        error(e)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_cdd(args):
    from pharmpy.tools.cdd.results import psn_cdd_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_cdd_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_frem(args):
    """Generate frem results"""
    from pharmpy.tools.frem.results import psn_frem_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_frem_results(
        args.psn_dir,
        method=args.method,
        force_posdef_covmatrix=args.force_posdef_covmatrix,
        force_posdef_samples=args.force_posdef_samples,
    )
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')

    from pharmpy.tools.reporting import create_report

    create_report(res, args.psn_dir)


def results_linearize(args):
    from pharmpy.tools.linearize.results import psn_linearize_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_linearize_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_resmod(args):
    from pharmpy.tools.ruvsearch.results import psn_resmod_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_resmod_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_scm(args):
    from pharmpy.tools.scm.results import psn_scm_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_scm_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_simeval(args):
    from pharmpy.tools.simeval.results import psn_simeval_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_simeval_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_ofv(args):
    """Subcommand to extract final ofv from multiple results"""
    ofvs = []
    for model in args.models:
        try:
            ofv = str(model.modelfit_results.ofv)
        except AttributeError:
            ofv = 'NA'
        ofvs.append(ofv)
    print(' '.join(ofvs))


def results_print(args):
    """Subcommand to print any results"""
    if args.dir.is_dir():
        path = args.dir / 'results.json'
    elif args.dir.is_file():
        path = args.dir
    else:
        error(FileNotFoundError(str(args.dir)))
    from pharmpy.workflows.results import read_results

    res = read_results(path)
    print(res)


def results_qa(args):
    from pharmpy.tools.qa.results import psn_qa_results

    if not args.psn_dir.is_dir():
        error(FileNotFoundError(str(args.psn_dir)))
    res = psn_qa_results(args.psn_dir)
    res.to_json(path=args.psn_dir / 'results.json')
    res.to_csv(path=args.psn_dir / 'results.csv')


def results_report(args):
    """Subcommand to generate reports"""
    results_path = args.psn_dir / 'results.json'
    if not results_path.is_file():
        error(FileNotFoundError(results_path))
    from pharmpy.workflows.results import read_results

    res = read_results(results_path)
    from pharmpy.tools.reporting import create_report

    create_report(res, args.psn_dir)


def results_summary(args):
    """Subcommand to output summary of modelfit"""
    for model in args.models:
        print(model.name)
        print(model.modelfit_results.parameter_summary())
        print()


def check_input_path(path):
    """Resolves path to input file and checks existence.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    try:
        path = Path(path)
        path = path_absolute(path)
    except FileNotFoundError:
        pass

    if not path.exists():
        exc = FileNotFoundError('No such input file: %r' % str(path))
        error(exc)
    elif path.is_dir():
        exc = IsADirectoryError('Is a directory (not an input file): %r' % str(path))
        error(exc)
    else:
        return path


# ------ Type functions --------------------


def input_model(path):
    """Returns :class:`~pharmpy.model.Model` from *path*.

    Raises if not found or is dir, without tracebacks (see :func:`error_exit`).
    """
    path = check_input_path(path)
    from pharmpy.model import Model
    from pharmpy.tools import read_modelfit_results

    model = Model.parse_model(path)
    res = read_modelfit_results(path)
    # FIXME: Should use tuple or something else instead
    model = model.replace(modelfit_results=res)
    return model


def input_model_or_dataset(path):
    """Returns :class:`~pharmpy.model.Model` or pd.DataFrame from *path*"""
    path = check_input_path(path)
    try:
        from pharmpy.model import Model

        obj = Model.parse_model(path)
    except TypeError:
        obj = pd.read_csv(path)
    return obj


def comma_list(s):
    if s == '':
        return []
    return s.split(',')


def semicolon_list(s):
    if s == '':
        return []
    return s.split(';')


def key_vals(a):
    d = {}
    for arg in a:
        pair = arg.split('=')
        d[pair[0]] = float(pair[1])
    return d


def random_seed(seed):
    try:
        import numpy as np

        seed = int(seed)
        np.random.seed(seed)
    except ValueError as e:
        # Cannot reraise the ValueError as it would be caught by argparse
        error(argparse.ArgumentTypeError(str(e)))
    return seed


# for commands taking multiple model files as input
args_input = argparse.ArgumentParser(add_help=False)
group_input = args_input.add_argument_group(title='inputs')
group_input.add_argument(
    'models', metavar='FILE', type=input_model, nargs='+', help='input model files'
)

# for commands taking one model file as input
args_model_input = argparse.ArgumentParser(add_help=False)
group_model_input = args_model_input.add_argument_group(title='model')
group_model_input.add_argument('model', metavar='FILE', type=input_model, help='input model file')

# for commands with model file or dataset input
args_model_or_data_input = argparse.ArgumentParser(add_help=False)
group_model_or_data_input = args_model_or_data_input.add_argument_group(title='data_input')
group_model_or_data_input.add_argument(
    'model_or_dataset',
    metavar='FILE',
    type=input_model_or_dataset,
    help='input model or dataset file',
)

# for commands involving randomization
args_random = argparse.ArgumentParser(add_help=False)
group_random = args_random.add_argument_group(title='random seed')
group_random.add_argument(
    '--seed',
    type=random_seed,
    metavar='INTEGER',
    help='Provide a random seed. The seed must be an integer ' 'between 0 and 2^32 - 1',
)

# for commands with file output
args_output = argparse.ArgumentParser(add_help=False)
group_output = args_output.add_argument_group(title='outputs')
group_output.add_argument(
    '-f', '--force', action='store_true', help='remove existing destination files (all)'
)
group_output.add_argument(
    '-o', '--output_file', dest='output_file', metavar='file', type=Path, help='output file'
)

parser_definition = [
    {
        'run': {
            'subs': [
                {
                    'execute': {
                        'help': 'Execute one or more model',
                        'func': run_execute,
                        'parents': [args_input],
                    }
                },
                {
                    'bootstrap': {
                        'help': 'Bootstrap',
                        'func': run_bootstrap,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--samples',
                                'type': int,
                                'help': 'Number of bootstrap datasets',
                            },
                        ],
                    }
                },
                {
                    'modelsearch': {
                        'help': 'Search for structural best model',
                        'func': run_modelsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': 'mfl',
                                'type': str,
                                'help': 'Search space to test',
                            },
                            {
                                'name': 'algorithm',
                                'type': str,
                                'help': 'Algorithm to use',
                            },
                            {
                                'name': '--rank_type',
                                'type': str,
                                'help': 'Name of function to use for ranking '
                                'candidates (default is bic).',
                                'default': 'bic',
                            },
                            {
                                'name': '--cutoff',
                                'type': float,
                                'help': 'Which selection criteria to rank models on',
                                'default': None,
                            },
                            {
                                'name': '--iiv_strategy',
                                'type': str,
                                'help': 'If/how IIV should be added to candidate models',
                                'default': 'absorption_delay',
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'iivsearch': {
                        'help': 'Search for best model IIV model',
                        'func': run_iivsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': 'algorithm',
                                'type': str,
                                'help': 'Algorithm to use',
                            },
                            {
                                'name': '--iiv_strategy',
                                'type': str,
                                'help': 'If/how IIV should be added to start model',
                                'default': 'no_add',
                            },
                            {
                                'name': '--rank_type',
                                'type': str,
                                'help': 'Name of function to use for ranking '
                                'candidates (default is bic).',
                                'default': 'bic',
                            },
                            {
                                'name': '--cutoff',
                                'type': float,
                                'help': 'Which selection criteria to rank models on',
                                'default': None,
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'iovsearch': {
                        'help': 'Search for best model IOV model',
                        'func': run_iovsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--column',
                                'type': str,
                                'help': 'Name of column in dataset to use as '
                                'occasion column (default is "OCC")',
                                'default': 'OCC',
                            },
                            {
                                'name': '--list_of_parameters',
                                'type': comma_list,
                                'help': 'List of parameters to test IOV on',
                                'default': None,
                            },
                            {
                                'name': '--rank_type',
                                'type': str,
                                'help': 'Name of function to use for ranking '
                                'candidates (default is bic).',
                                'default': 'bic',
                            },
                            {
                                'name': '--cutoff',
                                'type': float,
                                'help': 'Which selection criteria to rank models on',
                                'default': None,
                            },
                            {
                                'name': '--distribution',
                                'type': str,
                                'help': 'Which distribution added IOVs should have '
                                '(default is same-as-iiv)',
                                'default': 'same-as-iiv',
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'covsearch': {
                        'help': 'Identify covariates that explain some of the IIV',
                        'func': run_covsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--effects',
                                'type': str,
                                'help': 'The candidate effects to search through (required)',
                            },
                            {
                                'name': '--p_forward',
                                'type': float,
                                'help': 'The p-value threshold for forward '
                                'steps (default is `0.05`)',
                                'default': 0.05,
                            },
                            {
                                'name': '--p_backward',
                                'type': float,
                                'help': 'The p-value threshold for backward '
                                'steps (default is `0.01`)',
                                'default': 0.01,
                            },
                            {
                                'name': '--max_steps',
                                'type': int,
                                'help': 'The maximum number of search '
                                'algorithm steps to perform, or `-1` '
                                'for no maximum (default).',
                                'default': -1,
                            },
                            {
                                'name': '--algorithm',
                                'type': str,
                                'help': "The search algorithm to use (default "
                                "is `'scm-forward-then-backward'`)",
                                'default': 'scm-forward-then-backward',
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'ruvsearch': {
                        'help': 'Search for best residual error model',
                        'func': run_ruvsearch,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--groups',
                                'type': int,
                                'help': 'Number of groups for the time varying models',
                                'default': 4,
                            },
                            {
                                'name': '--p_value',
                                'type': float,
                                'help': 'p_value to use for the likelihood ratio test',
                                'default': 0.05,
                            },
                            {
                                'name': '--skip',
                                'type': comma_list,
                                'help': 'List of models to not test',
                                'default': None,
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'allometry': {
                        'help': 'Add allometric scaling',
                        'func': run_allometry,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': '--allometric_variable',
                                'type': str,
                                'help': 'Name of the allometric variable',
                                'default': 'WT',
                            },
                            {
                                'name': '--reference_value',
                                'type': float,
                                'help': 'Reference value for the allometric variable',
                                'default': 70.0,
                            },
                            {
                                'name': '--parameters',
                                'type': comma_list,
                                'help': 'List of parameters to apply scaling to',
                                'default': None,
                            },
                            {
                                'name': '--initials',
                                'type': comma_list,
                                'help': 'Initial estimates for the exponents',
                                'default': None,
                            },
                            {
                                'name': '--lower_bounds',
                                'type': comma_list,
                                'help': 'Lower bounds for the exponents',
                                'default': None,
                            },
                            {
                                'name': '--upper_bounds',
                                'type': comma_list,
                                'help': 'Upper bounds for the exponents',
                                'default': None,
                            },
                            {
                                'name': '--non_fixed',
                                'help': 'Should the exponents not be fixed',
                                'action': 'store_true',
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'estmethod': {
                        'help': 'Find best estimation method',
                        'func': run_estmethod,
                        'parents': [args_model_input],
                        'args': [
                            {
                                'name': 'algorithm',
                                'type': str,
                                'help': 'Algorithm to use',
                            },
                            {
                                'name': '--methods',
                                'type': str,
                                'default': None,
                                'help': 'List of methods to try, mark group of methods in single '
                                'quote separated by spaces. Supported are: FOCE, FO, IMP, '
                                'IMPMAP, ITS, SAEM, LAPLACE, BAYES, or \'all\'. Default is None.',
                            },
                            {
                                'name': '--solvers',
                                'type': str,
                                'default': None,
                                'help': 'List of solvers to try, mark group of methods in single '
                                'quote separated by spaces. Supported are: CVODES '
                                '(ADVAN14), DGEAR (ADVAN8), DVERK (ADVAN6), IDA '
                                '(ADVAN15), LSODA (ADVAN13) and LSODI (ADVAN9), or \'all\'. '
                                'Default is None.',
                            },
                            {
                                'name': '--path',
                                'type': Path,
                                'help': 'Path to output directory',
                            },
                        ],
                    }
                },
                {
                    'amd': {
                        'help': 'Use Automatic Model Development tool to select PK model',
                        'func': run_amd,
                        'args': [
                            {
                                'name': 'input_path',
                                'type': str,
                                'help': 'Path to output directory',
                            },
                            {
                                'name': '--modeltype',
                                'type': str,
                                'default': 'basic_pk',
                                'help': 'Type of model to build. Currently only "basic_pk"',
                            },
                            {
                                'name': '--administration',
                                'type': str,
                                'default': 'oral',
                                'help': 'Type of model to build. One of "iv", "oral" or "ivoral"',
                            },
                            {
                                'name': '--cl_init',
                                'type': float,
                                'default': 0.01,
                                'help': 'Initial estimate for the population clearance',
                            },
                            {
                                'name': '--vc_init',
                                'type': float,
                                'default': 0.1,
                                'help': 'Initial estimate for the central compartment population volume',
                            },
                            {
                                'name': '--mat_init',
                                'type': float,
                                'default': 0.01,
                                'help': 'Initial estimate for the mean absorption time (not for iv models)',
                            },
                            {
                                'name': '--search_space',
                                'type': str,
                                'default': None,
                                'help': 'MFL for search space for structural model',
                            },
                            {
                                'name': '--lloq_method',
                                'type': str,
                                'default': None,
                                'help': 'Method for how to remove LOQ data. See '
                                '`transform_blq` for list of available methods',
                            },
                            {
                                'name': '--lloq_limit',
                                'type': float,
                                'default': None,
                                'help': 'Lower limit of quantification. If None LLOQ column '
                                'from dataset will be used',
                            },
                            {
                                'name': '--order',
                                'type': comma_list,
                                'default': None,
                                'help': 'Runorder of components',
                            },
                            {
                                'name': '--allometric_variable',
                                'type': str,
                                'default': None,
                                'help': 'Variable to use for allometry',
                            },
                            {
                                'name': '--occasion',
                                'type': str,
                                'default': None,
                                'help': 'Name of occasion column',
                            },
                            {
                                'name': '--path',
                                'type': str,
                                'default': None,
                                'help': 'Path to run AMD in',
                            },
                            {
                                'name': '--resume',
                                'type': bool,
                                'default': True,
                                'help': 'Whether to allow resuming previous run',
                            },
                        ],
                    }
                },
            ],
            'help': 'Run a tool',
            'title': 'Pharmpy commands for running tools',
            'metavar': 'TOOL',
        }
    },
    {
        'data': {
            'subs': [
                {
                    'write': {
                        'help': 'Write dataset',
                        'parents': [args_model_input, args_output],
                        'func': data_write,
                        'description': 'Write a dataset from model as the model sees '
                        'it. For NM-TRAN models this means to filter all IGNORE and '
                        'ACCEPT statements in $DATA and to convert the dataset to '
                        'csv format.',
                    }
                },
                {
                    'print': {
                        'help': 'Print dataset to stdout',
                        'parents': [args_model_or_data_input],
                        'func': data_print,
                        'description': 'Print whole dataset or selected columns from model or csv '
                        'file via a pager to stdout. For NM-TRAN models the dataset will '
                        'be filtered first.',
                        'args': [
                            {
                                'name': '--columns',
                                'metavar': 'COLUMNS',
                                'type': comma_list,
                                'help': 'Select specific columns (default is all)',
                            }
                        ],
                    }
                },
                {
                    'filter': {
                        'help': 'Filter rows of dataset',
                        'description': 'Filter rows of dataset via expressions. '
                        'All rows matching all expressions will be kept.abs '
                        'A new model with the filtered dataset connected is created.',
                        'parents': [args_model_or_data_input, args_output],
                        'func': data_filter,
                        'args': [{'name': 'expressions', 'nargs': argparse.REMAINDER}],
                    }
                },
                {
                    'append': {
                        'help': 'Append column to dataset',
                        'description': 'Append a column to dataset given an assignment expression.'
                        'The expression can contain already present columns of the dataset.',
                        'parents': [args_model_or_data_input, args_output],
                        'func': data_append,
                        'args': [{'name': 'expression'}],
                    }
                },
                {
                    'resample': {
                        'help': 'Resample dataset',
                        'description': 'Bootstrap resample datasets'
                        'Multiple new models and datasets will be created',
                        'parents': [args_model_or_data_input, args_random],
                        'func': data_resample,
                        'args': [
                            {
                                'name': '--group',
                                'metavar': 'COLUMN',
                                'type': str,
                                'default': 'ID',
                                'help': 'Column to use for grouping (default is ID)',
                            },
                            {
                                'name': '--resamples',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': 1,
                                'help': 'Number of resampled datasets (default 1)',
                            },
                            {
                                'name': '--stratify',
                                'metavar': 'COLUMN',
                                'type': str,
                                'help': 'Column to use for stratification',
                            },
                            {
                                'name': '--replace',
                                'action': 'store_true',
                                'help': 'Sample with replacement (default is without)',
                            },
                            {
                                'name': '--sample_size',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': None,
                                'help': 'Number of groups to sample for each resample',
                            },
                        ],
                    }
                },
                {
                    'deidentify': {
                        'help': 'Deidentify dataset',
                        'description': 'Deidentify dataset by renumbering the id column and '
                        'changing dates.',
                        'parents': [args_output],
                        'func': data_deidentify,
                        'args': [
                            {
                                'name': 'dataset',
                                'metavar': 'FILE',
                                'type': str,
                                'help': 'A csv file dataset',
                            },
                            {
                                'name': '--idcol',
                                'metavar': 'COLUMN',
                                'type': str,
                                'default': 'ID',
                                'help': 'id column name (default ID)',
                            },
                            {
                                'name': '--datecols',
                                'type': str,
                                'metavar': 'COLUMNS',
                                'default': '',
                                'help': 'Comma separated list of date column names',
                            },
                        ],
                    }
                },
                {
                    'reference': {
                        'help': 'Set reference values in dataset',
                        'description': 'Set values in specific columns to provided reference values',
                        'parents': [args_model_input, args_output],
                        'func': data_reference,
                        'args': [
                            {
                                'name': 'colrefs',
                                'type': str,
                                'nargs': '+',
                                'metavar': 'COLNAME=VALUE',
                                'default': '',
                                'help': 'List of pairs of column names and reference values',
                            },
                        ],
                    }
                },
            ],
            'help': 'Data manipulations',
            'title': 'Pharmpy data commands',
            'metavar': 'ACTION',
        }
    },
    {'info': {'help': 'Show Pharmpy information', 'title': 'Pharmpy information', 'func': info}},
    {
        'model': {
            'subs': [
                {
                    'print': {
                        'help': 'Format and print model overview',
                        'description': 'Print an overview of a model.',
                        'func': model_print,
                        'parents': [args_input],
                        'args': [
                            {
                                'name': '--explicit-odes',
                                'action': 'store_true',
                                'help': 'Print the ODE system explicitly instead of as a '
                                'compartmental graph',
                            }
                        ],
                    }
                },
                {
                    'sample': {
                        'help': 'Sampling of parameter inits using uncertainty given by covariance '
                        'matrix',
                        'description': 'Sample parameter initial estimates using uncertainty given'
                        'by covariance matrix.',
                        'func': model_sample,
                        'parents': [args_model_input, args_random],
                        'args': [
                            {
                                'name': '--samples',
                                'metavar': 'NUMBER',
                                'type': int,
                                'default': 1,
                                'help': 'Number of sampled models',
                            }
                        ],
                    }
                },
                {
                    'add_cov_effect': {
                        'help': 'Adds covariate effect',
                        'description': 'Add covariate effect to a model parameter',
                        'func': add_covariate_effect,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {'name': 'param', 'type': str, 'help': 'Individual parameter'},
                            {'name': 'covariate', 'type': str, 'help': 'Covariate'},
                            {'name': 'effect', 'type': str, 'help': 'Type of covariate effect'},
                            {
                                'name': '--operation',
                                'type': str,
                                'default': '*',
                                'help': 'Whether effect should be added or multiplied',
                            },
                        ],
                    }
                },
                {
                    'absorption_rate': {
                        'help': 'Set absorption rate for a PK model',
                        'description': 'Change absorption rate of a PK model to either bolus, 0th '
                        'order, 1th order or sequential 0-order 1-order.',
                        'func': model_absorption_rate,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'rate',
                                'choices': ['bolus', 'ZO', 'FO', 'seq-ZO-FO'],
                                'type': str,
                                'help': 'Absorption rate',
                            },
                        ],
                    }
                },
                {
                    'elimination_rate': {
                        'help': 'Set elimination rate for a PK model',
                        'description': 'Change elimination rate of a PK model to either 0th '
                        'order, 1th order, Michaelis-Menten or combined 1-order-Michaelis-Menten.',
                        'func': model_elimination_rate,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'rate',
                                'choices': ['ZO', 'FO', 'MM', 'comb-FO-MM'],
                                'type': str,
                                'help': 'Eliminiation rate',
                            },
                        ],
                    }
                },
                {
                    'transit_compartments': {
                        'help': 'Set the number of transit compartments for a PK model',
                        'description': 'Set the number of transit compartments of a PK model',
                        'func': model_transit_compartments,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'n',
                                'type': int,
                                'help': 'Number of compartments',
                            },
                        ],
                    }
                },
                {
                    'peripheral_compartments': {
                        'help': 'Set the number of peripheral compartments for a PK model',
                        'description': 'Set the number of peripheral compartments of a PK model',
                        'func': model_peripheral_compartments,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'n',
                                'type': int,
                                'help': 'Number of compartments',
                            },
                        ],
                    }
                },
                {
                    'error': {
                        'help': 'Set error model',
                        'description': 'Set the error model to either additive, proportional or '
                        'combined',
                        'func': model_error,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'error_type',
                                'choices': ['none', 'additive', 'propotional', 'combined'],
                                'type': str,
                                'help': 'Type of error model',
                            },
                        ],
                    }
                },
                {
                    'boxcox': {
                        'help': 'Applies boxcox transformation',
                        'description': 'Apply boxcox transformation on selected/all etas',
                        'func': boxcox,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'tdist': {
                        'help': 'Applies t-distribution transformation',
                        'description': 'Apply t-distribution transformation on selected/all etas',
                        'func': tdist,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'john_draper': {
                        'help': 'Applies John Draper transformation',
                        'description': 'Apply John Draper transformation on selected/all etas',
                        'func': john_draper,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces',
                            },
                        ],
                    }
                },
                {
                    'add_iiv': {
                        'help': 'Adds IIVs',
                        'description': 'Add IIVs to a model parameter',
                        'func': add_iiv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {'name': 'param', 'type': str, 'help': 'Individual parameter'},
                            {'name': 'expression', 'type': str, 'help': 'Expression for added eta'},
                            {
                                'name': '--operation',
                                'type': str,
                                'default': '*',
                                'help': 'Whether effect should be added or multiplied',
                            },
                            {
                                'name': '--eta_name',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom name of new eta',
                            },
                        ],
                    }
                },
                {
                    'add_iov': {
                        'help': 'Adds IOVs',
                        'description': 'Add IOVs to a random variable or a specified model '
                        'parameter',
                        'func': add_iov,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': 'occ',
                                'type': str,
                                'help': 'Name of occasion column',
                            },
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas or parameters, mark group of etas in single '
                                'quote separated by spaces',
                            },
                            {
                                'name': '--eta_names',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom names of new etas. Must be equal to the '
                                'number of input etas times the number of categories for '
                                'occasion.',
                            },
                        ],
                    }
                },
                {
                    'create_joint_distribution': {
                        'help': 'Combine etas into joint distributions',
                        'description': 'Creates full or partial block structures of etas (IIV)',
                        'func': create_joint_distribution,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--etas',
                                'type': str,
                                'default': None,
                                'help': 'List of etas, mark group of etas in single quote '
                                'separated by spaces. To combine some etas provide list, '
                                'for all etas no arguments are needed',
                            },
                        ],
                    }
                },
                {
                    'iiv_on_ruv': {
                        'help': 'Applies IIV to RUVs',
                        'description': 'Applies IIV to RUVs by multiplying epsilons with '
                        'exponential etas',
                        'func': iiv_on_ruv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--eps',
                                'type': str,
                                'default': None,
                                'help': 'List of epsilons, mark group of epsilons in single quote '
                                'separated by spaces. To apply to all epsilons, omit this '
                                'argument.',
                            },
                            {
                                'name': '--same_eta',
                                'type': bool,
                                'default': True,
                                'help': 'whether all RUVs from input should use the same new ETA '
                                'or if one ETA should be created for each RUV.',
                            },
                            {
                                'name': '--eta_names',
                                'type': str,
                                'default': None,
                                'help': 'Optional custom names of new etas. Must be equal to the '
                                'number of input etas times the number of categories for '
                                'occasion.',
                            },
                        ],
                    }
                },
                {
                    'remove_iiv': {
                        'help': 'Removes IIVs',
                        'description': 'Removes all IIV omegas given a list with eta names '
                        'and/or parameter names',
                        'func': remove_iiv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--to_remove',
                                'type': str,
                                'default': None,
                                'help': 'List of etas and/or parameter name to remove. To remove '
                                'all etas, omit this argument.',
                            },
                        ],
                    }
                },
                {
                    'remove_iov': {
                        'help': 'Removes IOVs',
                        'description': 'Removes all IOV omegas',
                        'func': remove_iov,
                        'parents': [args_model_input, args_output],
                    }
                },
                {
                    'power_on_ruv': {
                        'help': 'Applies power effect to RUVs.',
                        'description': 'Applies a power effect to provided epsilons.',
                        'func': power_on_ruv,
                        'parents': [args_model_input, args_output],
                        'args': [
                            {
                                'name': '--eps',
                                'type': str,
                                'default': None,
                                'help': 'List of epsilons, mark group of epsilons in single quote '
                                'separated by spaces. To apply to all epsilons, omit this '
                                'argument.',
                            },
                        ],
                    }
                },
                {
                    'update_inits': {
                        'help': 'Update inits using modelfit results.',
                        'description': 'Update inits using modelfit results.',
                        'func': update_inits,
                        'parents': [args_model_input, args_output],
                    }
                },
            ],
            'help': 'Model manipulations',
            'title': 'Pharmpy model commands',
            'metavar': 'ACTION',
        }
    },
    {
        'results': {
            'subs': [
                {
                    'bootstrap': {
                        'help': 'Generate bootstrap results',
                        'description': 'Generate results from a PsN bootstrap run',
                        'func': results_bootstrap,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN bootstrap run directory',
                            }
                        ],
                    }
                },
                {
                    'cdd': {
                        'help': 'Generate cdd results',
                        'description': 'Generate results from a PsN cdd run',
                        'func': results_cdd,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN cdd run directory',
                            }
                        ],
                    }
                },
                {
                    'frem': {
                        'help': 'Generate FREM results',
                        'description': 'Generate results from a PsN frem run',
                        'func': results_frem,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN frem run directory',
                            },
                            {
                                'name': '--method',
                                'choices': ['cov_sampling', 'bipp'],
                                'default': 'cov_sampling',
                                'help': 'Method to use for uncertainty of covariate effects',
                            },
                            {
                                'name': '--force_posdef_covmatrix',
                                'action': 'store_true',
                                'help': 'Should covariance matrix be forced to become positive '
                                'definite',
                            },
                            {
                                'name': '--force_posdef_samples',
                                'type': int,
                                'default': 500,
                                'help': 'Number of sampling tries to do before starting to force '
                                'posdef',
                            },
                        ],
                    }
                },
                {
                    'linearize': {
                        'help': 'Generate linearize results',
                        'description': 'Generate results from a PsN linearize run',
                        'func': results_linearize,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN linearize run directory',
                            }
                        ],
                    }
                },
                {
                    'ofv': {
                        'help': 'Extract OFVs from model runs',
                        'description': 'Extract OFVs from one or more model runs',
                        'parents': [args_input],
                        'func': results_ofv,
                    }
                },
                {
                    'print': {
                        'help': 'Print results',
                        'description': 'Print results from PsN run to stdout',
                        'func': results_print,
                        'args': [
                            {
                                'name': 'dir',
                                'metavar': 'file or directory',
                                'type': Path,
                                'help': 'Path to directory containing results.json '
                                'or directly to json results file',
                            }
                        ],
                    }
                },
                {
                    'qa': {
                        'help': 'Generate qa results',
                        'description': 'Generate results from a PsN qa run',
                        'func': results_qa,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN qa run directory',
                            }
                        ],
                    }
                },
                {
                    'report': {
                        'help': 'Generate results report',
                        'description': 'Generate results report for PsN run (currently only frem)',
                        'func': results_report,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN run directory',
                            }
                        ],
                    }
                },
                {
                    'ruvsearch': {
                        'help': 'Generate ruvsearch results',
                        'description': 'Generate results from a PsN ruvsearch run',
                        'func': results_resmod,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN ruvsearch run directory',
                            }
                        ],
                    }
                },
                {
                    'scm': {
                        'help': 'Generate scm results',
                        'description': 'Generate results from a PsN scm run',
                        'func': results_scm,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN scm run directory',
                            }
                        ],
                    }
                },
                {
                    'simeval': {
                        'help': 'Generate simeval results',
                        'description': 'Generate results from a PsN simeval run',
                        'func': results_simeval,
                        'args': [
                            {
                                'name': 'psn_dir',
                                'metavar': 'PsN directory',
                                'type': Path,
                                'help': 'Path to PsN simeval run directory',
                            }
                        ],
                    }
                },
                {
                    'summary': {
                        'help': 'Modelfit summary',
                        'description': 'Print a summary of a model estimates to stdout.',
                        'parents': [args_input],
                        'func': results_summary,
                    }
                },
            ],
            'help': 'Result extraction and generation',
            'title': 'Pharmpy result generation commands',
            'metavar': 'ACTION',
        }
    },
]


def generate_parsers(parsers):
    for command in parser_definition:
        ((cmd_name, cmd_dict),) = command.items()
        cmd_parser = parsers.add_parser(
            cmd_name, allow_abbrev=True, help=cmd_dict['help'], formatter_class=formatter
        )
        if 'subs' in cmd_dict:
            subs = cmd_parser.add_subparsers(title=cmd_dict['title'], metavar=cmd_dict['metavar'])
            for sub_command in cmd_dict['subs']:
                ((sub_name, sub_dict),) = sub_command.items()
                args = sub_dict.pop('args', [])
                func = sub_dict.pop('func')

                sub_parser = subs.add_parser(sub_name, **sub_dict, formatter_class=formatter)
                for arg in args:
                    name = arg.pop('name')
                    sub_parser.add_argument(name, **arg)
                sub_parser.set_defaults(func=func)


parser = argparse.ArgumentParser(
    prog='pharmpy',
    description=dedent(
        """
    Welcome to the command line interface of Pharmpy!

    Functionality is split into various subcommands
        - try --help after a COMMAND
        - all keyword arguments can be abbreviated if unique


    """
    ).strip(),
    epilog=dedent(
        """
        Examples:
            # Create 100 bootstrap datasets
            pharmpy data resample pheno_real.mod --resamples=100 --replace

            # prettyprint model
            pharmpy model print pheno_real.mod

            # version/install information
            pharmpy info
    """
    ).strip(),
    formatter_class=formatter,
    allow_abbrev=True,
)
parser.add_argument('--version', action='version', version=pharmpy.__version__)

# subcommand parsers
subparsers = parser.add_subparsers(title='Pharmpy commands', metavar='COMMAND')
generate_parsers(subparsers)


# -- entry point of CLI (pharmpy) ---------------------------------------------------------


def main(args):
    """Entry point of ``pharmpy`` CLI util and ``python3 -m pharmpy`` (via ``__main__.py``)."""
    # parse
    args = parser.parse_args(args)
    # dispatch subcommand
    if 'func' in args:
        args.func(args)
    else:
        parser.print_usage()



import re


def detect_model(src):
    """Check if src represents a NONMEM control stream
    i.e. check if it is a file that contain $PRO
    """
    if not isinstance(src, str):
        return False
    is_control_stream = re.search(r'^\s*\$PRO', src, re.MULTILINE)
    return is_control_stream



from os.path import normpath, relpath
from pathlib import Path
from typing import Union


def path_relative_to(root: Path, path: Path) -> Path:
    # NOTE: A ValueError will be raised on Windows if path and root are on
    # different drives.
    return Path(normpath(relpath(str(path_absolute(path)), start=str(path_absolute(root)))))


def path_absolute(path: Path) -> Path:
    # NOTE: This makes the path absolute without resolving symlinks
    new_path = (
        Path(normpath(str(path))) if path.is_absolute() else Path(normpath(str(Path.cwd() / path)))
    )
    assert new_path.is_absolute()
    return new_path


def normalize_user_given_path(path: Union[str, Path]) -> Path:
    if isinstance(path, str):
        path = Path(path)
    return path.expanduser()



"""
NONMEM abbreviated record class.
"""

import re
from itertools import islice

from .record import Record


def _is_quoted_with(q: str, s: str):
    return s.startswith(q) and s.endswith(q)


def _is_quoted(s: str):
    return _is_quoted_with("'", s) or _is_quoted_with('"', s)


def strip_quote(s):
    return s[1:-1] if _is_quoted(s) else s


class AbbreviatedRecord(Record):
    @property
    def replace(self):
        """Give a dict of all REPLACE in record"""
        d = {}
        for replace in self.root.subtrees('replace'):
            a, b = islice(iter(replace.leaves('ANY')), 2)
            first = strip_quote(a.value)
            second = strip_quote(b.value)
            d[first] = second
        return d

    def translate_to_pharmpy_names(self):
        return {value: re.sub(r'\((\w+)\)', r'_\1', key) for key, value in self.replace.items()}



import re
from pathlib import Path

from .option_record import EnumOpt, IntOpt, MxOpt, OptionRecord, Opts, SimpleOpt, StrOpt, WildOpt

table_options = Opts(
    MxOpt('PRINT', 'print', default=True),
    MxOpt('NOPRINT', 'print'),
    StrOpt('FILE'),
    MxOpt('NOHEADER', 'header'),
    MxOpt('ONEHEADER', 'header'),
    SimpleOpt('ONEHEADERALL', noabbrev=True),
    MxOpt('NOTITLE', 'title'),
    MxOpt('NOLABEL', 'title'),
    MxOpt('FIRSTONLY', 'only'),
    MxOpt('LASTONLY', 'only'),
    MxOpt('FIRSTLASTONLY', 'only'),
    MxOpt('NOFORWARD', 'forward'),
    MxOpt('FORWARD', 'forward'),
    MxOpt('APPEND', 'append', default=True),
    MxOpt('NOAPPEND', 'append'),
    StrOpt('FORMAT', default='s1PE11.4'),
    StrOpt('LFORMAT'),
    StrOpt('RFORMAT'),
    StrOpt('IDFORMAT'),
    EnumOpt('NOSUB', (0, 1), default=0),
    StrOpt('PARAFILE'),
    IntOpt('ESAMPLE', default=300),
    SimpleOpt('WRESCHOL'),
    IntOpt('SEED', default=11456),
    EnumOpt('CLOCKSEED', (0, 1)),
    StrOpt('RANMETHOD'),
    EnumOpt('VARCALC', (0, 1, 2, 3)),
    StrOpt('FIXEDETAS'),
    EnumOpt('NPDTYPE', (0, 1), default=0),
    MxOpt('UNCONDITIONAL', 'cond', default=True),
    MxOpt('CONDITIONAL', 'cond'),
    SimpleOpt('OMITTED'),
    WildOpt(),
    # BY and EXCLUDE_BY missing
)


class TableRecord(OptionRecord):
    option_defs = table_options

    def __init__(self, name, raw_name, root):
        # Overriding to not parse options at parse time.
        # Need other records to be parsed first.
        super(OptionRecord, self).__init__(name, raw_name, root)

    @property
    def path(self):
        file_option = self.option_pairs['FILE']
        assert file_option is not None
        return Path(file_option)

    def set_path(self, value):
        newrec = self.set_option('FILE', str(value))
        return newrec

    @property
    def eta_derivatives(self):
        """List of numbers for etas whose derivative are requested"""
        return self._find_derivatives('G')

    @property
    def epsilon_derivatives(self):
        """List of numbers for epsilons whose derivateives are requested"""
        return self._find_derivatives('H')

    def _find_derivatives(self, ch):
        derivs = []
        regexp = ch + r'(\d+)1$'
        for key, value in self.all_options:
            m = re.match(regexp, key)
            if not m and value is not None:
                m = re.match(regexp, value)
            if m:
                n = m.group(1)
                derivs.append(int(n))
        return derivs

    def parse_options(self, nonoptions, netas):
        return self.option_defs.parse_ast(self.root, nonoptions=nonoptions, netas=netas)


# FIXME: These situations are not handled
# No columns are allowed after first option
# Overridden names are ignored after first option (not error)



from typing import Any, Dict, Literal, Optional

from pharmpy.model import EstimationStep, EstimationSteps, Model, SimulationStep
from pharmpy.modeling.help_functions import _as_integer

ESTIMATION_METHODS = ('FO', 'FOCE', 'ITS', 'LAPLACE', 'IMPMAP', 'IMP', 'SAEM', 'BAYES')


def set_estimation_step(model: Model, method: Literal[ESTIMATION_METHODS], idx: int = 0, **kwargs):
    """Set estimation step

    Sets estimation step for a model. Methods currently supported are:
        FO, FOCE, ITS, LAPLACE, IMPMAP, IMP, SAEM, BAYES

    Parameters
    ----------
    model : Model
        Pharmpy model
    method : {'FO', 'FOCE', 'ITS', 'LAPLACE', 'IMPMAP', 'IMP', 'SAEM', 'BAYES'}
        estimation method to change to
    idx : int
        index of estimation step, default is 0 (first estimation step)
    kwargs
        Arguments to pass to EstimationStep (such as interaction, evaluation)

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> opts = {'NITER': 1000, 'ISAMPLE': 100}
    >>> model = set_estimation_step(model, 'IMP', evaluation=True, tool_options=opts)
    >>> model.estimation_steps[0]   # doctest: +ELLIPSIS
    EstimationStep('IMP', interaction=True, parameter_uncertainty_method='SANDWICH', evaluation=True, ...,
    tool_options=...

    See also
    --------
    add_estimation_step
    remove_estimation_step
    append_estimation_step_options
    add_parameter_uncertainty_step
    remove_parameter_uncertainty_step
    set_evaluation_step

    """
    try:
        idx = _as_integer(idx)
    except TypeError:
        raise TypeError(f'Index must be integer: {idx}')

    d = kwargs
    d['method'] = method
    steps = model.estimation_steps
    newstep = steps[idx].replace(**d)
    newsteps = steps[0:idx] + newstep + steps[idx + 1 :]
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()


def add_estimation_step(
    model: Model, method: Literal[ESTIMATION_METHODS], idx: Optional[int] = None, **kwargs
):
    """Add estimation step

    Adds estimation step for a model in a given index. Methods currently supported are:
        FO, FOCE, ITS, LAPLACE, IMPMAP, IMP, SAEM

    Parameters
    ----------
    model : Model
        Pharmpy model
    method : {'FO', 'FOCE', 'ITS', 'LAPLACE', 'IMPMAP', 'IMP', 'SAEM', 'BAYES'}
        estimation method to change to
    idx : int
        index of estimation step (starting from 0), default is None (adds step at the end)
    kwargs
        Arguments to pass to EstimationStep (such as interaction, evaluation)

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> opts = {'NITER': 1000, 'ISAMPLE': 100}
    >>> model = add_estimation_step(model, 'IMP', tool_options=opts)
    >>> ests = model.estimation_steps
    >>> len(ests)
    2
    >>> ests[1]   # doctest: +ELLIPSIS
    EstimationStep('IMP', interaction=False, parameter_uncertainty_method=None, ..., tool_options={'NITER': 1000,...

    See also
    --------
    set_estimation_step
    remove_estimation_step
    append_estimation_step_options
    add_parameter_uncertainty_step
    remove_parameter_uncertainty_step
    set_evaluation_step

    """
    meth = EstimationStep.create(method, **kwargs)

    if idx is not None:
        try:
            idx = _as_integer(idx)
        except TypeError:
            raise TypeError(f'Index must be integer: {idx}')
        steps = model.estimation_steps
        newsteps = steps[0:idx] + meth + steps[idx:]
    else:
        newsteps = model.estimation_steps + meth
    model = model.replace(estimation_steps=newsteps)

    return model.update_source()


def set_simulation(model: Model, n: int = 1, seed: int = 64206):
    """Change model into simulation model

    Parameters
    ----------
    model : Model
        Pharmpy model
    n : int
        Number of replicates
    seed : int
        Random seed for the simulation

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_simulation(model, n=10, seed=1234)
    >>> steps = model.estimation_steps
    >>> steps[0]
    SimulationStep(n=10, seed=1234, solver=None, solver_rtol=None, solver_atol=None, tool_options=None)

    """
    final_est = None
    for step in model.estimation_steps:
        if isinstance(step, EstimationStep):
            final_est = step
    if final_est is not None:
        step = SimulationStep.create(
            n=n,
            seed=seed,
            solver=final_est.solver,
            solver_atol=final_est.solver_atol,
            solver_rtol=final_est.solver_rtol,
        )
    else:
        step = SimulationStep.create(n=n, seed=seed)
    steps = EstimationSteps((step,))
    model = model.replace(estimation_steps=steps)
    return model.update_source()


def remove_estimation_step(model: Model, idx: int):
    """Remove estimation step

    Parameters
    ----------
    model : Model
        Pharmpy model
    idx : int
        index of estimation step to remove (starting from 0)

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = remove_estimation_step(model, 0)
    >>> ests = model.estimation_steps
    >>> len(ests)
    0

    See also
    --------
    add_estimation_step
    set_estimation_step
    append_estimation_step_options
    add_parameter_uncertainty_step
    remove_parameter_uncertainty_step
    set_evaluation_step

    """
    try:
        idx = _as_integer(idx)
    except TypeError:
        raise TypeError(f'Index must be integer: {idx}')

    steps = model.estimation_steps
    newsteps = steps[0:idx] + steps[idx + 1 :]
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()


def append_estimation_step_options(model: Model, tool_options: Dict[str, Any], idx: int):
    """Append estimation step options

    Appends options to an existing estimation step.

    Parameters
    ----------
    model : Model
        Pharmpy model
    tool_options : dict
        any additional tool specific options
    idx : int
        index of estimation step (starting from 0)

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> opts = {'NITER': 1000, 'ISAMPLE': 100}
    >>> model = append_estimation_step_options(model, tool_options=opts, idx=0)
    >>> est = model.estimation_steps[0]
    >>> len(est.tool_options)
    2

    See also
    --------
    add_estimation_step
    set_estimation_step
    remove_estimation_step
    add_parameter_uncertainty_step
    remove_parameter_uncertainty_step
    set_evaluation_step

    """
    try:
        idx = _as_integer(idx)
    except TypeError:
        raise TypeError(f'Index must be integer: {idx}')

    steps = model.estimation_steps
    toolopts = dict(steps[idx].tool_options)
    toolopts.update(tool_options)
    newstep = steps[idx].replace(tool_options=toolopts)
    newsteps = steps[0:idx] + newstep + steps[idx + 1 :]
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()


def add_parameter_uncertainty_step(
    model: Model, parameter_uncertainty_method: Literal['SANDWICH', 'CPG', 'OFIM', 'EFIM']
):
    """Adds parameter uncertainty step to the final estimation step

    Parameters
    ----------
    model : Model
        Pharmpy model
    parameter_uncertainty_method : str
        Parameter uncertainty method to use

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_estimation_step(model, 'FOCE', parameter_uncertainty_method=None)
    >>> model = add_parameter_uncertainty_step(model, 'SANDWICH')
    >>> ests = model.estimation_steps
    >>> ests[0]   # doctest: +ELLIPSIS
    EstimationStep('FOCE', interaction=True, parameter_uncertainty_method='SANDWICH', ...)

    See also
    --------
    add_estimation_step
    set_estimation_step
    remove_estimation_step
    append_estimation_step_options
    remove_parameter_uncertainty_step
    set_evaluation_step

    """
    steps = model.estimation_steps
    newstep = steps[-1].replace(parameter_uncertainty_method=f'{parameter_uncertainty_method}')
    newsteps = steps[0:-1] + newstep
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()


def remove_parameter_uncertainty_step(model: Model):
    """Removes parameter uncertainty step from the final estimation step

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = remove_parameter_uncertainty_step(model)
    >>> ests = model.estimation_steps
    >>> ests[0]   # doctest: +ELLIPSIS
    EstimationStep('FOCE', interaction=True, parameter_uncertainty_method=None, ...)

    See also
    --------
    add_estimation_step
    set_estimation_step
    remove_estimation_step
    append_estimation_step_options
    add_parameter_uncertainty_step
    set_evaluation_step

    """
    steps = model.estimation_steps
    newstep = steps[-1].replace(parameter_uncertainty_method=None)
    newsteps = steps[:-1] + newstep
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()


def set_evaluation_step(model: Model, idx: int = -1):
    """Set estimation step

    Sets estimation step for a model. Methods currently supported are:
        FO, FOCE, ITS, LAPLACE, IMPMAP, IMP, SAEM, BAYES

    Parameters
    ----------
    model : Model
        Pharmpy model
    idx : int
        Index of estimation step, default is -1 (last estimation step)

    Returns
    -------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> model = set_evaluation_step(model)
    >>> model.estimation_steps[0]   # doctest: +ELLIPSIS
    EstimationStep('FOCE', interaction=True, parameter_uncertainty_method='SANDWICH', evaluation=True, ...

    See also
    --------
    set_estimation_step
    add_estimation_step
    remove_estimation_step
    append_estimation_step_options
    add_parameter_uncertainty_step
    remove_parameter_uncertainty_step

    """
    try:
        idx = _as_integer(idx)
    except TypeError:
        raise TypeError(f'Index must be integer: {idx}')

    steps = model.estimation_steps
    newstep = steps[idx].replace(evaluation=True)
    if idx != -1:
        newsteps = steps[0:idx] + newstep + steps[idx + 1 :]
    else:
        newsteps = steps[0:-1] + newstep
    model = model.replace(estimation_steps=newsteps)
    return model.update_source()



import io
import os
import re
import shutil
from contextlib import redirect_stdout

import pytest

import pharmpy.cli as cli
from pharmpy.internals.fs.cwd import chdir


def test_model_print(datadir, capsys):
    args = ['model', 'print', str(datadir / 'pheno.mod')]
    cli.main(args)
    captured = capsys.readouterr()
    assert 'ETA₁' in captured.out


@pytest.mark.parametrize('operation', ['*', '+'])
def test_add_covariate_effect(datadir, operation, tmp_path):
    shutil.copy(datadir / 'pheno.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')
    shutil.copy(datadir / 'pheno.datainfo', tmp_path / 'pheno.datainfo')

    with chdir(tmp_path):
        args = ['model', 'add_cov_effect', 'run1.mod', 'CL', 'WGT', 'exp', '--operation', operation]
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert not re.search('CLWGT', mod_ori)
        assert re.search('CLWGT', mod_cov)


@pytest.mark.parametrize(
    'transformation, eta', [('boxcox', 'ETAB1'), ('tdist', 'ETAT1'), ('john_draper', 'ETAD1')]
)
def test_eta_transformation(datadir, transformation, eta, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', transformation, 'run1.mod', '--etas', 'ETA_1']
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_box:
            mod_ori = f_ori.read()
            mod_box = f_box.read()

        assert mod_ori != mod_box

        assert not re.search(eta, mod_ori)
        assert re.search(eta, mod_box)


@pytest.mark.parametrize(
    'options, eta_name',
    [
        (['--operation', '+'], 'ETA_S1'),
        (['--operation', '*'], 'ETA_S1'),
        (['--eta_name', 'ETA(3)'], 'ETA(3)'),
    ],
)
def test_add_iiv(datadir, options, eta_name, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'add_iiv', 'run1.mod', 'S1', 'exp'] + options
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov
        assert f'EXP({eta_name})' not in mod_ori
        assert f'EXP({eta_name})' in mod_cov


@pytest.mark.parametrize('options', [['--eta_names', 'ETA(3) ETA(4)']])
def test_add_iov(datadir, options, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'add_iov', 'run1.mod', 'FA1', '--etas', 'ETA_1'] + options
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert not re.search(r'ETAI1', mod_ori)
        assert re.search(r'ETAI1', mod_cov)


def test_results_linearize(datadir, tmp_path):
    path = datadir / 'linearize' / 'linearize_dir1'
    shutil.copy(path / 'pheno_linbase.mod', tmp_path)
    shutil.copy(path / 'pheno_linbase.ext', tmp_path)
    shutil.copy(path / 'pheno_linbase.lst', tmp_path)
    shutil.copy(path / 'pheno_linbase.phi', tmp_path)
    scmdir = tmp_path / 'scm_dir1'
    scmdir.mkdir()
    shutil.copy(path / 'scm_dir1' / 'derivatives.mod', scmdir)
    shutil.copy(path / 'scm_dir1' / 'derivatives.ext', scmdir)
    shutil.copy(path / 'scm_dir1' / 'derivatives.lst', scmdir)
    shutil.copy(path / 'scm_dir1' / 'derivatives.phi', scmdir)

    args = ['results', 'linearize', str(tmp_path)]
    cli.main(args)

    assert os.path.exists(tmp_path / 'results.json')


@pytest.mark.parametrize('eta_args', [['--etas', 'ETA_1 ETA_2'], []])
def test_create_joint_distribution(datadir, eta_args, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'create_joint_distribution', 'run1.mod'] + eta_args
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert not re.search(r'BLOCK\(2\)', mod_ori)
        assert re.search(r'BLOCK\(2\)', mod_cov)


@pytest.mark.parametrize(
    'epsilons_args, eta_name',
    [
        (['--eps', 'EPS_1'], 'ETA_RV1'),
        ([], 'ETA_RV1'),
        (['--same_eta', 'False'], 'ETA_RV1'),
        (['--eta_names', 'ETA_3'], 'ETA(3)'),
    ],
)
def test_iiv_on_ruv(datadir, epsilons_args, eta_name, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'iiv_on_ruv', 'run1.mod'] + epsilons_args
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov
        assert f'EXP({eta_name})' not in mod_ori
        assert f'EXP({eta_name})' in mod_cov


@pytest.mark.parametrize('to_remove', [['--to_remove', 'ETA_2'], []])
def test_remove_iiv(datadir, to_remove, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'remove_iiv', 'run1.mod'] + to_remove
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert re.search(r'EXP\(ETA\(2\)\)', mod_ori)
        assert not re.search(r'EXP\(ETA\(2\)\)', mod_cov)


def test_remove_iov(datadir, tmp_path):
    shutil.copy(datadir / 'qa/iov.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'remove_iov', 'run1.mod']
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert re.search('SAME', mod_ori)
        assert not re.search('SAME', mod_cov)

        assert re.search(r'ETA\(3\)', mod_ori)
        assert not re.search(r'ETA\(3\)', mod_cov)


@pytest.mark.parametrize('epsilons_args', [['--eps', 'EPS_1'], []])
def test_power_on_ruv(datadir, epsilons_args, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'power_on_ruv', 'run1.mod'] + epsilons_args
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('run2.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

        assert mod_ori != mod_cov

        assert not re.search(r'\*\*', mod_ori)
        assert re.search(r'\*\*', mod_cov)


def test_update_inits(datadir, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno_real.ext', tmp_path / 'run1.ext')
    shutil.copy(datadir / 'pheno_real.phi', tmp_path / 'run1.phi')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'update_inits', 'run1.mod']
        cli.main(args)


def test_model_sample(datadir, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno_real.ext', tmp_path / 'run1.ext')
    shutil.copy(datadir / 'pheno_real.lst', tmp_path / 'run1.lst')
    shutil.copy(datadir / 'pheno_real.phi', tmp_path / 'run1.phi')
    shutil.copy(datadir / 'pheno_real.cov', tmp_path / 'run1.cov')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['model', 'sample', 'run1.mod', '--seed=24']
        cli.main(args)

        with open('run1.mod', 'r') as f_ori, open('sample_1.mod', 'r') as f_cov:
            mod_ori = f_ori.read()
            mod_cov = f_cov.read()

    assert mod_ori != mod_cov


def test_usage():
    f = io.StringIO()
    with redirect_stdout(f):
        cli.main([])
        out = f.getvalue()
    assert 'usage:' in out


def test_main():
    import pharmpy.__main__ as main

    main


def test_reference(datadir, tmp_path):
    shutil.copy(datadir / 'pheno_real.mod', tmp_path / 'run1.mod')
    shutil.copy(datadir / 'pheno.dta', tmp_path / 'pheno.dta')

    with chdir(tmp_path):
        args = ['data', 'reference', 'run1.mod', 'WGT=70', 'APGR=4', '--output', 'run2.mod']
        cli.main(args)

        with open('run2.mod', 'r') as f_new:
            mod_new = f_new.read()

        assert re.search(r'run2\.csv', mod_new)



"""The lock module exposes a path_lock context manager function

Exception classes are part of the public API.
Other exported functions are implementation details subject to change.

Caveats:
    If you need to lock multiple paths you should have a total order for each
    subset of paths that could be locked simultaneously by the same thread and call
    path_lock multiple times respecting this total order. For instance via
    resolved absolute path lexicographical order.

    The Linux implementation relies on fcntl/lockf which is infamous for being
    hard to work with: if you close any file descriptor for a given file in a
    given process, it will release the lock held by that process, even if the
    lock was acquired through a different file descriptor. So you cannot easily
    open a lock file for reading or writing without breaking the locking
    mechanism. We implement locks in a way that uses a single fd while a lock
    exists for a given (normalized) path. That fd is opened with O_RDWR and is
    yielded by path_lock for convenience, but the user should take extra care
    not to close this fd, as it would release the lock. For instance, `open`
    must be used with the `closefd` flag as in `open(fd, closefd=False)`. This
    makes writing to the lock file, or reading several times from the lock
    file, a bit challenging but not impossible (using `fp.seek` and
    `fp.truncate`).

    Another solution would be to use flock but since that can sometimes
    fallback to the fcntl/lockf implementation, we prefer to use the latter
    implementation directly, and workaround the limitations. See
    http://0pointer.de/blog/projects/locking.html.

    Currently, on Windows, at the process level, shared locks are "simulated" by
    exclusive locks. Implementing true shared locks on this platform would require
    depending on userland libraries. This would be necessary, for instance, in
    situations where multiple processes need to lock shared resources and
    communicate simultaneously about the state of those resources. Note that the
    problem in this particular example can be circumvented at the cost of
    efficiency, by simulating locked shared resources through multiple rounds of
    exclusively-locked resources plus communication. I do not think our current
    usage mandates the correct implementation of shared locks. It is enough that
    the implementation is simple for Windows, and is correct on UNIX.

    The current implementation may not even be correct on UNIX in the scenario
    discussed above: in a multi-process, multi-thread setting, if one thread of
    process A acquires an exlusive lock at the thread-level first, then
    attempts to exclusively acquire the process-level lock, that process-level
    lock may already be acquired as shared by a thread of another process B
    that is waiting on the other threads of A to run. And vice versa, if one
    thread of process A acquires an exclusive lock at the process-level first,
    then attempts to exclusively acquire the thread-level lock, that
    thread-level lock may already be acquired as shared by other threads of
    process A, but they are waiting on a thread of process B that is in turn
    waiting for the process-level lock to be downgraded to shared. So it seems
    that composing two isolated implementations, one for threads, one for
    processes, cannot work in these "share and communicate" scenarios. One
    solution is to forbid upgrade and downgrade of a process-level lock which
    would effectively be the same as having an all-or-nothing lock (either you
    lock at both levels or you do not).

    Currently, on Windows, at the process level, we attempt to lock the entire file
    by passing the largest possible value to `mscvrt.locking` `nbytes` argument. If
    we want this implementation to be correct, we would need to `fo.seek(0)` before
    calling `mscvrt.locking` then perhaps restore the seek position? Also the
    largest possible value is 2^31-1 which might not be enough for files larger
    than 2GB. Again, we do not currently need this "functionality".

    It is currently not possible to specify a timeout for acquiring a lock. The
    only options currently are immediate failure or forever blocking.
"""
import os
import sys
from collections import Counter
from contextlib import contextmanager
from threading import Condition, Lock, RLock, get_ident
from typing import Generic, TypeVar

T = TypeVar('T')


class AcquiringLockWouldBlockError(Exception):
    """Raised when acquiring lock would block and blocking is False"""


class AcquiringProcessLevelLockWouldBlockError(AcquiringLockWouldBlockError):
    """Raised when acquiring a process-level lock would block and blocking is False"""


class AcquiringThreadLevelLockWouldBlockError(AcquiringLockWouldBlockError):
    """Raised when acquiring a thread-level lock would block and blocking is False"""


class RecursiveDeadlockError(Exception):
    """Raised when recursive dead-lock is detected."""


is_windows = os.name == 'nt'
is_mac_os = sys.platform == 'darwin'

if is_windows:
    # Windows file locking
    import msvcrt
    import sys

    # NOTE: lock the entire file
    _lock_length = -1 if sys.version_info.major == 2 else int(2**31 - 1)

    def _is_process_level_lock_timeout_error(error: OSError) -> bool:
        """Check if an OSError corresponds to a blocking lock timeout error

        Note that the errno for a non-blocking lock failure would be 13 with a
        strerror of "Permission denied".
        """
        return error.errno == 36 and error.strerror == 'Resource deadlock avoided'

    def _is_process_level_lock_blocking_error(error: OSError) -> bool:
        """Check that an OSError corresponds to an error raised because
        a process-level lock cannot be acquired immediately
        """
        return (
            isinstance(error, PermissionError)
            and error.errno == 13
            and error.strerror == 'Permission denied'
        )

    def _process_level_lock(fd: int, shared: bool = False, blocking: bool = True):
        # NOTE: Simulates shared lock using an exclusive lock. This
        # implementation does not allow to lock the same fd multiple times.
        # This does not matter a we make sure we do not do that.
        if blocking:
            while True:
                try:
                    msvcrt.locking(  # pyright: ignore [reportGeneralTypeIssues]
                        fd,
                        msvcrt.LK_LOCK,  # pyright: ignore [reportGeneralTypeIssues]
                        _lock_length,
                    )
                    break
                except OSError as error:
                    if not _is_process_level_lock_timeout_error(error):
                        raise error
        else:
            try:
                msvcrt.locking(  # pyright: ignore [reportGeneralTypeIssues]
                    fd, msvcrt.LK_NBLCK, _lock_length  # pyright: ignore [reportGeneralTypeIssues]
                )
            except PermissionError as error:
                if _is_process_level_lock_blocking_error(error):
                    raise AcquiringProcessLevelLockWouldBlockError()
                else:
                    raise error

    def _process_level_unlock(fd: int):
        # NOTE: This implementation (Windows) will raise an error if attempting
        # to unlock an already unlocked fd. This does not matter as we make
        # sure we do not do that.
        msvcrt.locking(  # pyright: ignore [reportGeneralTypeIssues]
            fd, msvcrt.LK_UNLCK, _lock_length  # pyright: ignore [reportGeneralTypeIssues]
        )

else:
    # UNIX based file locking
    import fcntl

    def _is_process_level_lock_blocking_error(error: OSError) -> bool:
        """Check that an OSError corresponds to an error raised because
        a process-level lock cannot be acquired immediately
        """
        return (
            isinstance(error, BlockingIOError)
            and (
                (is_mac_os and error.errno == 35)  # MacOS
                or (not (is_mac_os) and error.errno == 11)  # Linux
            )
            and error.strerror == 'Resource temporarily unavailable'
        )

    def _process_level_lock(fd: int, shared: bool = False, blocking: bool = True):
        operation = fcntl.LOCK_SH if shared else fcntl.LOCK_EX
        if blocking:
            fcntl.lockf(fd, operation)
        else:
            try:
                fcntl.lockf(fd, operation | fcntl.LOCK_NB)
            except BlockingIOError as error:
                if _is_process_level_lock_blocking_error(error):
                    raise AcquiringProcessLevelLockWouldBlockError()
                else:
                    raise error

    def _process_level_unlock(fd: int):
        # NOTE: This implementation (UNIX) will NOT raise an error if attempting
        # to unlock an already unlocked fd. This does not matter as we make
        # sure we do not do that.
        fcntl.lockf(fd, fcntl.LOCK_UN)


class ShareableProcessLock:
    """Creates a process lock for a FD shared by all threads of a process"""

    def __init__(self, fd: int):
        self._fd = fd
        self._lock = Lock()
        self._shared_by: Counter[int] = Counter()
        self._exclusively_held_by: Counter[int] = Counter()

    @contextmanager
    def lock(self, shared: bool = False, blocking: bool = True, reentrant: bool = False):
        """Locks the scoped FD

        shared : bool
            Whether the lock should be shared. If not shared, the lock is
            exclusive. Exclusively locking a file only means that we want to
            exclude other processes from locking the same file: any number of
            threads of the same process can lock the same file both shared and
            exclusively simultaneously. We keep track of which threads are
            holding what kind of locks to upgrade the process lock from shared
            to exclusive and vice versa.
        blocking : bool
            Whether lock acquisition should be blocking. If True, will raise an
            error if a lock cannot be acquired immediately. Otherwise, will block
            until the lock can be acquired.
        reentrant : bool
            Whether lock acquisition is reentrant. If True, allows each thread
            to lock the file recursively. Otherwise, the same thread
            recursively locking dead-locks.
        """
        thread_id = get_ident()
        if self._lock.acquire(blocking=blocking):
            try:
                if not reentrant and (
                    self._shared_by[thread_id] or self._exclusively_held_by[thread_id]
                ):
                    raise RecursiveDeadlockError()

                is_held_shared = bool(self._shared_by)
                is_held_exclusively = bool(self._exclusively_held_by)
                is_held = is_held_shared or is_held_exclusively

                if not is_held or (is_held_shared and not shared and not is_windows):
                    # NOTE: We only lock when the first locking attempt is made, or
                    # if we want to upgrade the lock to an exclusive lock, but only on
                    # UNIX since current implementation always uses exclusive locks on
                    # Windows.
                    _process_level_lock(self._fd, shared, blocking)

                if shared:
                    self._shared_by[thread_id] += 1
                else:
                    self._exclusively_held_by[thread_id] += 1
            finally:
                self._lock.release()
        else:
            raise AcquiringProcessLevelLockWouldBlockError()

        try:
            yield

        finally:
            with self._lock:
                if shared:
                    self._shared_by[thread_id] -= 1
                    if not self._shared_by[thread_id]:
                        del self._shared_by[thread_id]
                else:
                    self._exclusively_held_by[thread_id] -= 1
                    if not self._exclusively_held_by[thread_id]:
                        del self._exclusively_held_by[thread_id]

                is_held_shared = bool(self._shared_by)
                is_held_exclusively = bool(self._exclusively_held_by)
                is_held = is_held_shared or is_held_exclusively

                if not is_held:
                    # NOTE: We only release the lock once we have exhausted all
                    # locking attempts
                    _process_level_unlock(self._fd)
                elif not is_held_exclusively and not shared and not is_windows:
                    # NOTE: We downgrade the lock from exclusive to shared if we
                    # are not on Windows and we just released an exclusive
                    # lock, and no exclusive lock is left.
                    _process_level_lock(self._fd, shared=True, blocking=True)


class ShareableThreadLock:
    def __init__(self):
        self._condition = Condition(RLock())
        self._acquired_by: Counter[int] = Counter()

    def lock(self, shared: bool = False, blocking: bool = True, reentrant: bool = False):
        if shared:
            return self._lock_sh(blocking=blocking, reentrant=reentrant)
        else:
            return self._lock_ex(blocking=blocking, reentrant=reentrant)

    @contextmanager
    def _lock_sh(self, blocking: bool = True, reentrant: bool = False):
        thread_id = get_ident()
        if self._condition.acquire(blocking=blocking):
            try:
                if not reentrant and self._acquired_by[thread_id]:
                    raise RecursiveDeadlockError()
                self._acquired_by[thread_id] += 1
            finally:
                self._condition.release()
        else:
            raise AcquiringThreadLevelLockWouldBlockError()

        try:
            yield

        finally:
            self._condition.acquire(blocking=True)
            try:
                self._acquired_by[thread_id] -= 1
                if not self._acquired_by[thread_id]:
                    del self._acquired_by[thread_id]  # NOTE: GC
                    if not self._acquired_by:
                        self._condition.notify_all()
            finally:
                self._condition.release()

    @contextmanager
    def _lock_ex(self, blocking: bool = True, reentrant: bool = False):
        thread_id = get_ident()
        if self._condition.acquire(blocking=blocking):
            acquired = False
            try:
                this_thread_count = Counter({thread_id: self._acquired_by[thread_id]})
                if blocking:
                    # NOTE: We could use self._acquired_by != this_thread_count in
                    # Python >= 3.10
                    while self._acquired_by - this_thread_count:
                        self._condition.wait()
                else:
                    # NOTE: We could use self._acquired_by != this_thread_count in
                    # Python >= 3.10
                    if self._acquired_by - this_thread_count:
                        raise AcquiringThreadLevelLockWouldBlockError()

                # NOTE: The following check also works for Python < 3.10 because we
                # systematically get rid of zero-count entries.
                if self._acquired_by:
                    # NOTE: The following two checks can be replaced by
                    # assert self._acquired_by == this_thread_count
                    # in Python >= 3.10
                    assert len(self._acquired_by) == 1
                    assert self._acquired_by[thread_id] == this_thread_count[thread_id]
                    if not reentrant:
                        raise RecursiveDeadlockError()

                acquired = True
                self._acquired_by[thread_id] += 1

                yield

            finally:
                if acquired:
                    self._acquired_by[thread_id] -= 1
                    if not self._acquired_by[thread_id]:
                        del self._acquired_by[thread_id]  # NOTE: GC
                self._condition.release()
        else:
            raise AcquiringThreadLevelLockWouldBlockError()


class ThreadSafeKeyedRefPool(Generic[T]):
    def __init__(self, lock, refs: dict, factory, destructor=None):
        self._lock = lock
        self._refs = refs
        self._factory = factory
        self._destructor = destructor

    @contextmanager
    def __call__(self, key: T):
        with self._lock:
            entry = self._refs.get(key)
            if entry is None:
                # NOTE: We create a new object if none exists
                obj = self._factory(key)
                self._refs[key] = (obj, 1)
            else:
                # NOTE: Otherwise we count one ref more
                (obj, refcount) = entry
                self._refs[key] = (obj, refcount + 1)

        try:
            yield obj

        finally:
            with self._lock:
                (_, refcount) = self._refs[key]
                if refcount == 1:
                    # NOTE: We remove the object from the pool since nobody
                    # else holds a reference to it.
                    del self._refs[key]
                    if self._destructor is not None:
                        self._destructor(obj)
                else:
                    # NOTE: Otherwise we count one ref less
                    self._refs[key] = (obj, refcount - 1)


_thread_level_lock_ref = ThreadSafeKeyedRefPool(Lock(), {}, lambda _key: ShareableThreadLock())
_process_level_lock_ref = ThreadSafeKeyedRefPool(Lock(), {}, ShareableProcessLock)
_fd_ref = ThreadSafeKeyedRefPool(
    Lock(), {}, lambda normalized_path: os.open(normalized_path, os.O_RDWR), lambda fd: os.close(fd)
)


@contextmanager
def process_level_lock(
    fd: int, shared: bool = False, blocking: bool = True, reentrant: bool = False
):
    with _process_level_lock_ref(fd) as ref:
        with ref.lock(shared, blocking, reentrant):
            yield


@contextmanager
def thread_level_lock(
    key: str, shared: bool = False, blocking: bool = True, reentrant: bool = False
):
    with _thread_level_lock_ref(key) as ref:
        with ref.lock(shared, blocking, reentrant):
            yield


@contextmanager
def process_level_path_lock(
    normalized_path: str, shared: bool = False, blocking: bool = True, reentrant: bool = False
):
    with _fd_ref(normalized_path) as fd:
        with process_level_lock(fd, shared, blocking, reentrant):
            yield fd


@contextmanager
def path_lock(path: str, shared: bool = False, blocking: bool = True, reentrant: bool = False):
    """Locks a path both at the thread-level and process-level.

    Parameters
    ----------
    path : str
        The path to lock. This path will be open in read-only mode if shared is
        True, write mode otherwise. Must be the path of an existing file (not a
        directory).
    shared : bool
        Whether the lock should be shared. If not shared, the lock is
        exclusive. One can have either a single exclusive lock or any number of
        shared lock on the same resource at any given time.
    blocking : bool
        Whether lock acquisition should be blocking. If True, will raise an
        error if a lock cannot be acquired immediately. Otherwise, will block
        until the lock can be acquired.
    reentrant : bool
        Whether lock acquisition is reentrant. If True, allows to lock
        recursively. Otherwise, recursively locking dead-locks.
    """
    key = os.path.normpath(path)
    with thread_level_lock(key, shared, blocking, reentrant):
        with process_level_path_lock(key, shared, blocking, reentrant) as fd:
            yield fd



from pharmpy.model import Model


def get_lag_times(model: Model):
    """Get lag times for all compartments

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    dict
        Dictionary from compartment name to lag time expression
    """
    odes = model.statements.ode_system
    assert odes is not None
    names = odes.compartment_names
    d = {}
    for name in names:
        cmt = odes.find_compartment(name)
        if cmt.lag_time != 0:
            d[name] = cmt.lag_time
    return d


def get_bioavailability(model: Model):
    """Get bioavailability of doses for all compartments

    Parameters
    ----------
    model : Model
        Pharmpy model

    Returns
    -------
    dict
        Dictionary from compartment name to bioavailability expression
    """
    odes = model.statements.ode_system
    assert odes is not None
    names = odes.compartment_names
    d = {}
    for name in names:
        cmt = odes.find_compartment(name)
        if cmt.bioavailability != 1:
            d[name] = cmt.bioavailability
    return d



"""
NONMEM simulation record class.
"""

from typing import cast

from pharmpy.internals.parse.generic import eval_token

from .record import Record


class SimulationRecord(Record):
    @property
    def nsubs(self):
        """Number of subproblems"""
        n = cast(int, eval_token(self.root.subtree('nsubs').leaf('INT')))
        # NOTE: According to NONMEM documentation 0 means 1 here
        return 1 if n == 0 else n



from pharmpy.internals.immutable import frozenmapping
from pharmpy.modeling import convert_model, create_basic_pk_model, set_zero_order_absorption


def test_create_basic_pk_model(testdata):
    model = create_basic_pk_model('iv')
    assert len(model.parameters) == 6

    model = create_basic_pk_model('oral')
    assert len(model.parameters) == 8

    dataset_path = testdata / 'nonmem/pheno.dta'
    model = create_basic_pk_model(
        administration='oral', dataset_path=dataset_path, cl_init=0.01, vc_init=1.0, mat_init=0.1
    )
    model = convert_model(model, 'nonmem')
    run1 = set_zero_order_absorption(model)
    assert not model.dataset.empty
    assert not run1.dataset.empty
    assert isinstance(model.dependent_variables, frozenmapping)

    dataset_path = testdata / 'nonmem/pheno_pd.csv'
    model = create_basic_pk_model(
        administration='oral', dataset_path=dataset_path, cl_init=0.01, vc_init=1.0, mat_init=0.1
    )
    assert not model.dataset.empty
    assert model.datainfo._separator == r'\s+'

    model = create_basic_pk_model('ivoral')
    assert len(model.statements.ode_system.dosing_compartments) == 2



"""
The NONMEM $MODEL record
"""

from .option_record import OptionRecord


class ModelRecord(OptionRecord):
    @property
    def ncomps(self):
        nc = self.get_option("NCOMPARTMENTS")
        if nc is None:
            nc = self.get_option("NCOMPS")
            if nc is None:
                nc = self.get_option("NCM")
        if nc is not None:
            nc = int(nc)
        return nc

    def add_compartment(self, name, dosing=False):
        options = (name, 'DEFDOSE') if dosing else (name,)
        newrec = self.append_option('COMPARTMENT', f'({" ".join(options)})')
        return newrec

    def get_compartment_number(self, name):
        for i, (curname, _) in enumerate(self.compartments()):
            if name == curname:
                return i + 1
        return None

    def compartments(self):
        ncomps = self.ncomps
        if (
            ncomps is not None
            and not self.has_option("COMPARTMENT")
            and not self.has_option("COMP")
        ):
            for i in range(1, ncomps + 1):
                yield f'COMP{i}', []
            return

        all_options = [
            'INITIALOFF',
            'NOOFF',
            'NODOSE',
            'EQUILIBRIUM',
            'EXCLUDE',
            'DEFOBSERVATION',
            'DEFDOSE',
        ]
        for n, opts in enumerate(self.get_option_lists('COMPARTMENT')):
            name = f'COMP{n + 1}'
            options = []
            for opt in opts:
                match = OptionRecord.match_option(all_options, opt)
                if match:
                    options.append(match)
                else:
                    name = opt
            yield name, options



import warnings
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Any, Dict, Optional, Union

import pharmpy.visualization
from pharmpy.deps import numpy as np
from pharmpy.deps import pandas as pd
from pharmpy.deps.scipy import stats
from pharmpy.model import Model
from pharmpy.tools import read_modelfit_results
from pharmpy.tools.psn_helpers import cmd_line_model_path, model_paths
from pharmpy.workflows.results import ModelfitResults, Results


@dataclass(frozen=True)
class BootstrapResults(Results):
    rst_path = Path(__file__).resolve().parent / 'report.rst'

    # FIXME: Should inherit from results that take multiple runs like bootstrap, cdd etc.
    parameter_statistics: Optional[Any] = None
    parameter_distribution: Optional[Any] = None
    covariance_matrix: Optional[Any] = None
    ofv_distribution: Optional[Any] = None
    ofv_statistics: Optional[Any] = None
    included_individuals: Optional[Any] = None
    ofvs: Optional[Any] = None
    parameter_estimates: Optional[Any] = None
    ofv_plot: Optional[Any] = None
    parameter_estimates_correlation_plot: Optional[Any] = None
    dofv_quantiles_plot: Optional[Any] = None
    parameter_estimates_histogram: Optional[Any] = None


def plot_ofv(res):
    plot = pharmpy.visualization.histogram(
        res.ofvs['bootstrap_bootdata_ofv'], title='Bootstrap OFV'
    )
    return plot


def plot_dofv_quantiles(res):
    ofvs = res.ofvs
    dofvs = ofvs['delta_bootdata'].sort_values().reset_index(drop=True)
    dofvs_boot_base = ofvs['delta_origdata'].sort_values().reset_index(drop=True)
    quantiles = np.linspace(0.0, 1.0, num=len(dofvs))
    degrees = len(res.parameter_distribution)
    chi2_dist = stats.chi2(df=degrees)
    chi2 = chi2_dist.ppf(quantiles)
    degrees_dofvs = res.ofv_statistics['mean']['delta_bootdata']
    degrees_boot_base = res.ofv_statistics['mean']['delta_origdata']
    df_dict: Dict[Union[str, tuple], Any] = {
        'quantiles': quantiles,
        f'Reference χ²({degrees})': chi2,
    }
    if not np.isnan(degrees_dofvs):
        df_dict[
            (
                'Original model OFV - Bootstrap model OFV (both using bootstrap datasets)',
                f'Estimated df = {degrees_dofvs:.2f}',
            )
        ] = dofvs
    if not np.isnan(degrees_boot_base):
        df_dict[
            (
                'Bootstrap model OFV - Original model OFV (both using original dataset)',
                f'Estimated df = {degrees_boot_base:.2f}',
            )
        ] = dofvs_boot_base
    df = pd.DataFrame(df_dict)
    plot = pharmpy.visualization.line_plot(
        df,
        'quantiles',
        xlabel='Distribution quantiles',
        ylabel='dOFV',
        legend_title='Distribution',
        title='dOFV quantiles',
    )
    return plot


def plot_parameter_estimates_correlation(res):
    pe = res.parameter_estimates
    plot = pharmpy.visualization.scatter_matrix(pe)
    return plot


def plot_parameter_estimates_histogram(res):
    pe = res.parameter_estimates
    plot = pharmpy.visualization.facetted_histogram(pe)
    return plot


def calculate_results(
    bootstrap_models, results, original_results=None, included_individuals=None, dofv_results=None
):
    if original_results is None:
        warnings.warn(
            'No results for the base model could be read. Cannot calculate bias and '
            'original_bootdata_ofv'
        )

    df = pd.DataFrame()
    for res in results:
        df = pd.concat([df, res.parameter_estimates], axis=1, ignore_index=True, sort=False)
    df = df.T
    df = df.reindex(results[0].parameter_estimates.index, axis=1)
    parameter_estimates = df.reset_index(drop=True)

    covariance_matrix = df.cov()

    df = parameter_estimates.copy()
    mean = df.mean()
    if original_results is not None:
        orig = original_results.parameter_estimates
        bias = mean - orig
    else:
        bias = np.nan
    statistics = pd.DataFrame(
        {'mean': mean, 'median': df.median(), 'bias': bias, 'stderr': df.std()}
    )
    statistics['RSE'] = statistics['stderr'] / statistics['mean']

    df = parameter_estimates
    distribution = create_distribution(df)

    boot_ofvs = [x.ofv for x in results]
    ofvs = pd.Series(boot_ofvs, name='bootstrap_bootdata_ofv')
    ofvs = pd.DataFrame(ofvs)
    ofvs['original_bootdata_ofv'] = np.nan
    ofvs['bootstrap_origdata_ofv'] = np.nan
    if original_results is not None:
        base_iofv = original_results.individual_ofv
        if included_individuals and base_iofv is not None:
            for i, included in enumerate(included_individuals):
                ofvs.at[i, 'original_bootdata_ofv'] = base_iofv[included].sum()

    if dofv_results is not None:
        for i, res in enumerate(dofv_results):
            if res is not None:
                ofvs.at[i, 'bootstrap_origdata_ofv'] = res.ofv

    if original_results is not None:
        ofvs['original_origdata_ofv'] = original_results.ofv

    ofvs['delta_bootdata'] = ofvs['original_bootdata_ofv'] - ofvs['bootstrap_bootdata_ofv']

    if original_results is not None:
        ofvs['delta_origdata'] = ofvs['bootstrap_origdata_ofv'] - ofvs['original_origdata_ofv']
    else:
        ofvs['delta_origdata'] = np.nan

    with warnings.catch_warnings():
        # Catch numpy warnings beause of NaN in ofvs
        warnings.filterwarnings('ignore', r'All-NaN slice encountered')
        warnings.filterwarnings('ignore', 'Mean of empty slice')
        ofv_dist = create_distribution(ofvs)
        ofv_stats = pd.DataFrame(
            {'mean': ofvs.mean(), 'median': ofvs.median(), 'stderr': ofvs.std()}
        )

    res = BootstrapResults(
        covariance_matrix=covariance_matrix,
        parameter_statistics=statistics,
        parameter_distribution=distribution,
        ofv_distribution=ofv_dist,
        ofv_statistics=ofv_stats,
        included_individuals=included_individuals,
        ofvs=ofvs,
        parameter_estimates=parameter_estimates,
    )

    return replace(
        res,
        ofv_plot=plot_ofv(res),
        # FIXME: Plot broken
        # parameter_estimates_correlation_plot = plot_parameter_estimates_correlation(res)
        dofv_quantiles_plot=plot_dofv_quantiles(res),
        parameter_estimates_histogram=plot_parameter_estimates_histogram(res),
    )


def create_distribution(df):
    dist = pd.DataFrame(
        {
            'min': df.min(),
            '0.05%': df.quantile(0.0005),
            '0.5%': df.quantile(0.005),
            '2.5%': df.quantile(0.025),
            '5%': df.quantile(0.05),
            'median': df.median(),
            '95%': df.quantile(0.95),
            '97.5%': df.quantile(0.975),
            '99.5%': df.quantile(0.995),
            '99.95%': df.quantile(0.9995),
            'max': df.max(),
        }
    )
    return dist


def psn_bootstrap_results(path):
    """Create bootstrapresults from a PsN bootstrap run

    :param path: Path to PsN boostrap run directory
    :return: A :class:`BootstrapResults` object
    """
    path = Path(path)

    paths = list(model_paths(path, 'bs_pr1_*.mod'))
    models = [Model.parse_model(p) for p in paths]
    # Read the results already now to give an appropriate error if no results exists
    results = []
    for p in paths:
        res = read_modelfit_results(p)
        if res is not None:
            results.append(res)
    if not results:
        raise FileNotFoundError("No model results available in m1")
    try:
        base_results = read_modelfit_results(cmd_line_model_path(path))
    except FileNotFoundError:
        base_results = None

    # Read dOFV results in NONMEM specific way. Models have multiple $PROBLEM
    # Create proper result objects to pass to calculate_results
    dofv_results = None
    if (path / 'm1' / 'dofv_1.mod').is_file():
        # FIXME: We should not depend on NONMEM here
        from pharmpy.model.external.nonmem.table import ExtTable, NONMEMTableFile

        dofv_results = []
        for table_path in (path / 'm1').glob('dofv_*.ext'):
            table_file = NONMEMTableFile(table_path)
            next_table = 1
            for table in table_file:
                assert isinstance(table, ExtTable)
                while next_table != table.number:
                    dofv_results.append(None)
                    next_table += 1
                res = ModelfitResults(ofv=table.final_ofv)
                dofv_results.append(res)
                next_table += 1

    incinds = pd.read_csv(path / 'included_individuals1.csv', header=None).values.tolist()
    res = calculate_results(
        models,
        results,
        original_results=base_results,
        included_individuals=incinds,
        dofv_results=dofv_results,
    )
    return res



import pytest

from pharmpy.model import EstimationStep, EstimationSteps, SimulationStep


def test_init():
    a = EstimationStep.create('foce', tool_options={'opt1': 23})
    assert a.tool_options['opt1'] == 23
    a = EstimationStep.create('foce', solver='lsoda')
    assert a.solver == 'LSODA'
    with pytest.raises(ValueError):
        EstimationStep.create('foce', solver='unknownsolverz')
    steps = EstimationSteps.create()
    assert len(steps) == 0


def test_estimation_method():
    a = EstimationStep.create('foce', parameter_uncertainty_method='sandwich')
    assert a.method == 'FOCE'
    assert a.parameter_uncertainty_method == 'SANDWICH'

    with pytest.raises(ValueError):
        EstimationStep.create('sklarfs')

    a = a.replace(method='fo')
    assert a.method == 'FO'
    assert a.parameter_uncertainty_method == 'SANDWICH'

    assert a == EstimationStep.create(
        'fo', interaction=False, parameter_uncertainty_method='sandwich'
    )
    assert (
        repr(a)
        == "EstimationStep('FO', interaction=False, parameter_uncertainty_method='SANDWICH', evaluation=False, "
        "maximum_evaluations=None, laplace=False, isample=None, niter=None, auto=None, "
        "keep_every_nth_iter=None, solver=None, solver_rtol=None, solver_atol=None, "
        "tool_options={})"
    )

    with pytest.raises(ValueError):
        EstimationStep.create('fo', maximum_evaluations=0)


def test_repr():
    steps = EstimationSteps()
    assert steps._repr_html_() == 'EstimationSteps()'
    assert repr(steps) == 'EstimationSteps()'

    a = EstimationStep.create('foce')
    steps = EstimationSteps.create([a])
    assert type(steps._repr_html_()) == str
    assert type(repr(steps)) == str


def test_eq():
    s1 = EstimationSteps()
    s2 = EstimationSteps()
    assert s1 == s2
    a = EstimationStep.create('foce')
    s3 = EstimationSteps.create([a, a])
    assert s1 != s3
    b = EstimationStep.create('fo')
    s4 = EstimationSteps.create([a, b])
    assert s3 != s4


def test_len():
    s1 = EstimationSteps()
    assert len(s1) == 0
    a = EstimationStep.create('foce')
    s2 = EstimationSteps.create([a, a])
    assert len(s2) == 2


def test_radd():
    a = EstimationStep.create('foce')
    s2 = EstimationSteps.create([a, a])
    b = EstimationStep.create('fo')
    conc = b + s2
    assert len(conc) == 3
    conc = (b,) + s2
    assert len(conc) == 3


def test_add():
    a = EstimationStep.create('foce')
    s2 = EstimationSteps.create([a, a])
    b = EstimationStep.create('fo')
    s3 = EstimationSteps.create([a, b])
    conc = s2 + b
    assert len(conc) == 3
    conc = s2 + (b,)
    assert len(conc) == 3
    conc = s2 + s3
    assert len(conc) == 4


def test_hash():
    a = EstimationStep.create('foce')
    b = EstimationStep.create('fo')
    assert hash(a) != hash(b)
    s1 = EstimationSteps.create([a, b])
    s2 = EstimationSteps.create([a])
    assert hash(s1) != hash(s2)


def test_dict():
    a = EstimationStep.create('foce')
    d = a.to_dict()
    assert d == {
        'method': 'FOCE',
        'interaction': False,
        'parameter_uncertainty_method': None,
        'evaluation': False,
        'maximum_evaluations': None,
        'laplace': False,
        'isample': None,
        'niter': None,
        'auto': None,
        'class': 'EstimationStep',
        'keep_every_nth_iter': None,
        'solver': None,
        'solver_rtol': None,
        'solver_atol': None,
        'tool_options': {},
    }
    step2 = EstimationStep.from_dict(d)
    assert step2 == a

    b = EstimationStep.create('fo')
    s1 = EstimationSteps.create([a, b])
    d = s1.to_dict()
    assert d == {
        'steps': (
            {
                'method': 'FOCE',
                'interaction': False,
                'parameter_uncertainty_method': None,
                'evaluation': False,
                'maximum_evaluations': None,
                'laplace': False,
                'isample': None,
                'niter': None,
                'auto': None,
                'class': 'EstimationStep',
                'keep_every_nth_iter': None,
                'solver': None,
                'solver_rtol': None,
                'solver_atol': None,
                'tool_options': {},
            },
            {
                'method': 'FO',
                'interaction': False,
                'parameter_uncertainty_method': None,
                'evaluation': False,
                'maximum_evaluations': None,
                'laplace': False,
                'isample': None,
                'niter': None,
                'auto': None,
                'class': 'EstimationStep',
                'keep_every_nth_iter': None,
                'solver': None,
                'solver_rtol': None,
                'solver_atol': None,
                'tool_options': {},
            },
        )
    }
    s2 = EstimationSteps.from_dict(d)
    assert s1 == s2


def test_getitem():
    a = EstimationStep.create('foce')
    b = EstimationStep.create('fo')
    s = EstimationSteps.create([a, b])
    assert s[0].method == 'FOCE'
    assert s[1].method == 'FO'

    assert len(s[1:]) == 1


def test_properties():
    a = EstimationStep.create('foce', epsilon_derivatives=['EPS(1)'])
    assert a.epsilon_derivatives == ('EPS(1)',)
    a = EstimationStep.create('foce', eta_derivatives=['ETA(1)'])
    assert a.eta_derivatives == ('ETA(1)',)
    a = EstimationStep.create('foce', predictions=['PRED'])
    assert a.predictions == ('PRED',)
    a = EstimationStep.create('foce', residuals=['CWRES'])
    assert a.residuals == ('CWRES',)


def test_replace():
    a = EstimationStep.create('foce')
    b = a.replace(method='fo')
    assert b.method == 'FO'
    c = a.replace(solver_atol=0.01)
    assert c.solver_atol == 0.01

    steps1 = EstimationSteps((a,))
    steps2 = EstimationSteps((b,))
    steps3 = steps1.replace(steps=[steps2])
    assert len(steps3) == 1


def test_simulation_step():
    ss = SimulationStep(n=23)
    assert ss.n == 23

    with pytest.raises(ValueError):
        SimulationStep.create(n=0)

    ss = SimulationStep.create(n=2)
    assert ss.n == 2
    ss = ss.replace(n=19)
    assert ss.n == 19

    ss = SimulationStep(seed=1234)
    assert ss.seed == 1234

    ss1 = SimulationStep()
    ss2 = SimulationStep(n=2)
    assert ss1 != ss2
    assert ss1 == ss1
    assert hash(ss1) != hash(ss2)

    d = ss2.to_dict()
    assert d == {
        'class': 'SimulationStep',
        'n': 2,
        'seed': 64206,
        'solver': None,
        'solver_atol': None,
        'solver_rtol': None,
        'tool_options': None,
    }
    assert ss2 == SimulationStep.from_dict(d)

    assert (
        repr(ss2)
        == 'SimulationStep(n=2, seed=64206, solver=None, solver_rtol=None, solver_atol=None, tool_options=None)'
    )



from dataclasses import dataclass
from typing import Literal, Tuple, Union

from lark.visitors import Interpreter

from .feature import ModelFeature, feature
from .symbols import Name, Wildcard


@dataclass(frozen=True)
class Elimination(ModelFeature):
    modes: Union[Tuple[Name[Literal['FO', 'ZO', 'MM', 'MIX-FO-MM']], ...], Wildcard]

    def __add__(self, other):
        if isinstance(self.modes, Wildcard) or isinstance(other.modes, Wildcard):
            return Elimination(Wildcard())
        else:
            return Elimination(self.modes + tuple([a for a in other.modes if a not in self.modes]))

    def __sub__(self, other):
        if isinstance(other.modes, Wildcard):
            return Elimination((Name('INST')))
        elif isinstance(self.modes, Wildcard):
            default = self._wildcard
            all_modes = tuple([a for a in default if a not in other.modes])
        else:
            # NOTE : WILDCARD should not be used here to future proof the method
            all_modes = tuple([a for a in self.modes if a not in other.modes])

        if len(all_modes) == 0:
            all_modes = (Name('FO'),)

        return Elimination(all_modes)

    def __eq__(self, other):
        if isinstance(other, Elimination):
            return set(self.modes) == set(other.modes)
        else:
            return False

    @property
    def eval(self):
        if isinstance(self.modes, Wildcard):
            return Elimination(self._wildcard)
        else:
            return self

    @property
    def _wildcard(self):
        return tuple([Name(x) for x in ['FO', 'ZO', 'MM', 'MIX-FO-MM']])


class EliminationInterpreter(Interpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        return feature(Elimination, children)

    def elimination_modes(self, tree):
        children = self.visit_children(tree)
        return list(Name(child.value.upper()) for child in children)

    def elimination_wildcard(self, tree):
        return Wildcard()
